# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Timeline base plot."""
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import attr
import pandas as pd
from bokeh.io import output_notebook, show
from bokeh.layouts import column
from bokeh.models import ColumnDataSource, HoverTool, LayoutDOM, Legend  # type: ignore
from bokeh.models.annotations import LegendItem
from bokeh.plotting import figure, reset_output

from .._version import VERSION
from ..common.data_utils import ensure_df_datetimes
from ..common.utility import check_kwargs, export
from .figure_dimension import bokeh_figure
from .timeline_common import (
    calc_auto_plot_height,
    create_range_tool,
    create_tool_tips,
    get_color_palette,
    get_def_source_cols,
    get_ref_event_time,
    get_tick_formatter,
    get_time_bounds,
    plot_ref_events,
    plot_ref_line,
    set_axes_and_grids,
)

# pylint: disable=unused-import
# Importing to activate pandas accessors
from .timeline_pd_accessor import TimeLineAccessor  # noqa F401
from .timeline_values import display_timeline_values  # noqa F401

# pylint: enable=unused-import

__version__ = VERSION
__author__ = "Ian Hellen"

# wrap figure function to handle v2/v3 parameter renaming
figure = bokeh_figure(figure)  # type: ignore[assignment, misc]


@attr.s(auto_attribs=True)
class PlotParams:
    """Plot params for time_duration."""

    time_column: str = "TimeGenerated"
    height: Optional[int] = None
    width: int = 900
    title: str = "Events"
    yaxis: bool = True
    range_tool: bool = True
    group_by: Optional[str] = None
    legend: Optional[str] = None
    xgrid: bool = True
    ygrid: bool = False
    hide: bool = False
    color: str = "navy"
    size: int = 10
    ylabel_cols: Iterable[str] = attr.Factory(list)
    ref_event: Optional[Any] = None
    ref_time: Optional[datetime] = None
    ref_events: Optional[pd.DataFrame] = None
    ref_col: Optional[str] = None
    ref_time_col: Optional[str] = None
    ref_times: Optional[List[Tuple[datetime, str]]] = None
    ref_label: str = "Ref time"
    source_columns: List[str] = []
    alert: Any = None
    overlay_color: Optional[str] = None
    overlay_data: Optional[pd.DataFrame] = None
    overlay_columns: Iterable[str] = attr.Factory(list)

    @classmethod
    def field_list(cls) -> List[str]:
        """Return field names as a list."""
        return list(attr.fields_dict(cls).keys())

    @property
    def fmt_title(self):
        """Return formatted title."""
        return f"Timeline: {self.title}"


@export
def display_timeline(
    data: Union[pd.DataFrame, dict],
    time_column: str = "TimeGenerated",
    source_columns: Optional[List[str]] = None,
    **kwargs,
) -> LayoutDOM:
    """
    Display a timeline of events.

    Parameters
    ----------
    data : Union[dict, pd.DataFrame]
        Either
        dict of data sets to plot on the timeline with the following structure::

            Key (str) - Name of data set to be displayed in legend
            Value (Dict[str, Any]) - containing:
                data (pd.DataFrame) - Data to plot
                time_column (str, optional) - Name of the timestamp column
                source_columns (list[str], optional) - source columns to use
                    in tooltips
                color (str, optional) - color of data points for this data
                size (int) - size of plotted event glyphs
            If any of the last values are omitted, they default to the values
            supplied as parameters to the function (see below)

        Or
        DataFrame as a single data set or grouped into individual
        plot series using the `group_by` parameter
    time_column : str, optional
        Name of the timestamp column
        (the default is 'TimeGenerated')
    source_columns : Optional[List[str]]
        List of default source columns to use in tooltips
        (the default is None)

    Other Parameters
    ----------------
    title : str, optional
        Title to display (the default is None)
    alert : SecurityAlert, optional
        Add a reference line/label using the alert time (the default is None)
    ref_event : Any, optional
        Add a reference line/label using the alert time (the default is None)
    ref_time : datetime, optional
        Add a reference line/label using `ref_time` (the default is None)
    group_by : str
        (where `data` is a DataFrame)
        The column to group timelines on
    size : Union[int, str]
        The size of the event glyph.
        If a string the size is taken as a column in the input data.
        If an integer, this is used as the fixed size.
    legend: str, optional
        "left", "right", "inline" or "none"
        (the default is to show a legend when plotting multiple series
        and not to show one when plotting a single series)
    yaxis : bool, optional
        Whether to show the yaxis and labels (default is False)
    ygrid : bool, optional
        Whether to show the yaxis grid (default is False)
    xgrid : bool, optional
        Whether to show the xaxis grid (default is True)
    range_tool : bool, optional
        Show the the range slider tool (default is True)
    height : int, optional
        The height of the plot figure
        (the default is auto-calculated height)
    width : int, optional
        The width of the plot figure (the default is 900)
    color : str
        Default series color (default is "navy")
    overlay_data : pd.DataFrame:
        A second dataframe to plot as a different series.
    overlay_color : str
        Overlay series color (default is "green")
    hide : bool, optional
        If True, create but do not display the plot.
        By default, False.
    ref_events : pd.DataFrame, optional
        Add references line/label using the event times in the dataframe.
        (the default is None)
    ref_time_col : str, optional
        Add references line/label using the this column in `ref_events`
        for the time value (x-axis).
        (this defaults the value of the `time_column` parameter or 'TimeGenerated'
        `time_column` is None)
    ref_col : str, optional
        The column name to use for the label from `ref_events`
        (the default is None)
    ref_times : List[Tuple[datetime, str]], optional
        Add one or more reference line/label using (the default is None)

    Returns
    -------
    LayoutDOM
        The bokeh plot figure.

    """
    # Get args
    check_kwargs(kwargs, PlotParams.field_list())
    param = PlotParams(
        time_column=time_column, source_columns=source_columns or [], **kwargs
    )
    param.ref_time, param.ref_label = get_ref_event_time(**kwargs)

    if isinstance(data, pd.DataFrame):
        if param.overlay_data is not None:
            aggr_data = {
                "Primary": {
                    "data": data,
                    "time_column": param.time_column,
                    "source_columns": source_columns,
                    "color": param.color,
                    "size": param.size,
                },
                "Secondary": {
                    "data": param.overlay_data,
                    "time_column": time_column,
                    "source_columns": param.overlay_columns,
                    "color": param.overlay_color,
                    "size": param.size,
                },
            }
            return _display_timeline_dict(data=aggr_data, param=param)

        # Create a dictionary from a grouped or simple series
        series_dict = _create_dict_from_grouping(
            data=data,
            source_columns=param.source_columns,
            time_column=param.time_column,
            group_by=param.group_by,
            color=param.color,
            size=param.size,
        )
        return _display_timeline_dict(data=series_dict, param=param)

    if isinstance(data, dict):
        return _display_timeline_dict(data, param=param)
    raise TypeError(
        "`data` parameter must be a DataFrame or a Python dict.",
        f"The type supplied was {type(data)}",
    )


def _display_timeline_dict(
    data: dict, param: PlotParams
) -> LayoutDOM:  # noqa: C901, MC0001
    """
    Display a timeline of events.

    Parameters
    ----------
    data : dict
        Data points to plot on the timeline.
        The dict should contain the following structure:
            Key - Name of data type to be displayed in legend
            Value - dict of data containing:
                data : pd.DataFrame
                    Data to plot
                time_column : str
                    Name of the timestamp column
                source_columns : list
                    List of source columns to use in tooltips
                color: str
                    Color of data points for this data
    param : PlotParams
        Plot parameters object

    Returns
    -------
    LayoutDOM
        The bokeh plot layout.

    """
    reset_output()
    output_notebook()

    tool_tip_columns, min_time, max_time = _unpack_data_series_dict(data, param)
    if min_time is None or max_time is None:
        print("No data to plot.")
        return figure()

    series_count = len(data)

    hover = HoverTool(**(create_tool_tips(data, tool_tip_columns)))

    start_range, end_range, min_time, max_time = get_time_bounds(min_time, max_time)
    height = param.height or calc_auto_plot_height(len(data))
    y_range = ((-1 / series_count), series_count - 1 + (1 / series_count))
    plot = figure(
        x_range=(start_range, end_range),
        y_range=y_range,
        min_border_left=50,
        x_axis_label="Event Time",
        x_axis_type="datetime",
        x_minor_ticks=10,
        tools=[hover, "xwheel_zoom", "box_zoom", "reset", "save", "xpan"],
        title=param.fmt_title,
        height=height,
        width=param.width,
    )

    set_axes_and_grids(data, plot, param.yaxis, param.ygrid, param.xgrid)  # type: ignore

    # Create plot bar to act as as range selector
    rng_select = create_range_tool(
        data=data,  # type: ignore
        min_time=min_time,
        max_time=max_time,
        plot_range=plot.x_range,  # type: ignore[arg-type]
        width=param.width,
        height=height,
    )

    # set the tick datetime formatter
    plot.xaxis[0].formatter = get_tick_formatter()
    # plot the data
    _plot_series(data, plot, param.legend)

    if param.ref_time is not None:
        plot_ref_line(plot, param.ref_time, param.ref_label, len(data))  # type: ignore
    elif param.ref_events is not None or param.ref_times is not None:
        plot_ref_events(
            plot=plot,
            ref_events=param.ref_events,
            time_col=param.ref_time_col or param.time_column,
            group_count=series_count,
            ref_col=param.ref_col,
            ref_times=param.ref_times,
        )

    plot_layout = column(plot, rng_select) if param.range_tool else plot
    if not param.hide:
        show(plot_layout)

    return plot_layout


def _plot_series(data, plot, legend_pos):
    """Plot data series and add legend."""
    # plot groups individually so that we can create an interactive legend
    # if legend_pos is "inline", we add add the normal legend inside the plot
    # if legend_pos is "left" or "right", we add the legend to the side
    if len(data) > 1 and not legend_pos:
        legend_pos = "left"
    legend_items = []
    for ser_name, series_def in data.items():
        size_param = series_def.get("size", 10)
        glyph_size: Union[pd.Series, int]
        if isinstance(size_param, str):
            if size_param in series_def["data"].columns:
                glyph_size = series_def["data"][size_param]
            else:
                glyph_size = 10
        else:
            glyph_size = size_param
        time_col = "TimeGenerated"
        if "time_column" in series_def:
            time_col = series_def["time_column"]
        if legend_pos == "inline":
            p_series = plot.diamond(
                x=time_col,
                y="y_index",
                color=series_def["color"],
                alpha=0.5,
                size=glyph_size,
                source=series_def["source"],
                legend_label=str(ser_name),
            )
        else:
            p_series = plot.diamond(
                x=time_col,
                y="y_index",
                color=series_def["color"],
                alpha=0.5,
                size=glyph_size,
                source=series_def["source"],
            )
        if legend_pos in ["left", "right"]:
            legend_items.append(
                LegendItem(
                    label=str(ser_name),
                    renderers=[p_series],
                )
            )

    if legend_pos == "inline":
        # Position the inline legend
        plot.legend.location = "center_left"
        plot.legend.click_policy = "hide"
    elif legend_pos in ["left", "right"]:
        # Create the legend box outside of the plot area
        ext_legend = Legend(
            items=legend_items[::-1],  # the legend is in the wrong order otherwise
            location="center",
            click_policy="hide",
            label_text_font_size="8pt",
        )
        plot.add_layout(ext_legend, legend_pos)


# pylint: disable=too-many-locals
def _unpack_data_series_dict(data, param: PlotParams):
    """Unpack each series from the data series dictionary."""
    # Process the input dictionary
    # Take each item that is passed and fill in blanks and add a y_index
    tool_tip_columns: Set[str] = set()
    min_time = None
    max_time = None
    y_index = 0

    # Create a color map in case colors have not been specified
    # (Shift the Viridis palette so we lose the top, harder-to-see colors)
    series_count = len(data)
    colors, palette_size = get_color_palette(series_count)

    for ser_name, series_def in data.items():
        data_columns: Set[str] = set()
        series_data = series_def["data"]

        if (
            series_data is None
            or not isinstance(series_data, pd.DataFrame)
            or series_data.empty
        ):
            print(f"No data to plot for series {ser_name}.")
            continue

        # if the series has source columns, use those
        # or fall back to global source cols or defaults
        series_def_src_cols = get_def_source_cols(series_data, param.source_columns)
        src_cols = series_def.get("source_columns", series_def_src_cols)
        data_columns.update(src_cols or series_def_src_cols)

        time_col = series_def.get("time_column", None)
        if not time_col:
            time_col = param.time_column
            series_def["time_column"] = time_col
        # ensure named time columns are in datetime format
        series_data = ensure_df_datetimes(data=series_data, columns=time_col)
        data_columns.update([time_col])
        # add the data columns to the tool tip column set
        tool_tip_columns.update(data_columns)

        # Create the Column data source to plot
        graph_df = series_data[list(data_columns)].copy()
        graph_df["y_index"] = y_index
        series_def["y_index"] = y_index
        ser_color = series_def.get("color", None)
        if not ser_color:
            ser_color = colors[y_index % palette_size]
            series_def["color"] = ser_color
        # Wrap tooltip lines longer than 50 chars
        _wrap_df_columns(graph_df, 50)
        series_def["source"] = ColumnDataSource(graph_df)
        y_index += 1

        # calculate min/max time from this set
        if min_time is None:
            min_time = series_data[time_col].min()
            max_time = series_data[time_col].max()
        else:
            min_time = min(min_time, series_data[time_col].min())
            max_time = max(max_time, series_data[time_col].max())

    return tool_tip_columns, min_time, max_time


# pylint: enable=too-many-locals


def _create_dict_from_grouping(
    data, source_columns, time_column, group_by, color, size=10
):
    """Return data groupings as a dictionary."""
    data_columns = get_def_source_cols(data, source_columns)
    # If the time column not explicitly specified in source_columns, add it
    data_columns.add(time_column)

    series_dict: Dict[str, Dict] = {}
    # create group frame so that we can color each group separately
    if group_by:
        data_columns.add(group_by)
        grouped_data = data[list(data_columns)].groupby(group_by)

        series_count = len(grouped_data)
        colors, palette_size = get_color_palette(series_count)
        for color_index, (group_name, group_df) in enumerate(grouped_data):
            series_dict[str(group_name)] = {
                "data": group_df,
                "time_column": time_column,
                "source_columns": source_columns,
                "color": colors[color_index % palette_size],
                "size": size,
            }
    else:
        group_df = data[list(data_columns)].copy()
        series_dict["unnamed series"] = {
            "data": group_df,
            "time_column": time_column,
            "source_columns": source_columns,
            "color": color,
            "size": size,
        }

    return series_dict


def _wrap_df_columns(data: pd.DataFrame, wrap_len: int = 50):
    """Wrap any string columns."""
    if not data.empty:
        for col in data.columns:
            if isinstance(data[col].iloc[0], str):
                data[col] = data[col].str.wrap(wrap_len)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Service Now Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import datetime as dt
from typing import Any, Dict, Tuple

import attr

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from .http_context_provider import HttpContextProvider

__version__ = VERSION
__author__ = "Florian Bracq"


_DEF_HEADERS = {
    "Content-Type": "application/json",
    "Accept": "application/json",
    "Accept-Charset": "UTF-8",
    "cache-control": "no-cache",
}


# pylint: disable=too-few-public-methods
@attr.s
class _ServiceNowParams(APILookupParams):
    # override LookupParams to set common defaults
    def __attrs_post_init__(self):
        self.auth_str = ["{ApiID}", "{AuthKey}"]
        self.auth_type = "HTTPBasic"


@export
class ServiceNow(HttpContextProvider):
    """ServiceNow Lookup."""

    _BASE_URL = "https://{Instance}.service-now.com/api/now/table"

    _SERVICE_NOW_PARAMS = {
        "sysparm_display_value": True,
        "sysparm_exclude_reference_link": True,
        "sysparm_limit": 10,
    }

    _QUERIES = {
        "ipv4": _ServiceNowParams(
            path="/cmdb_ci_computer",
            params={
                **_SERVICE_NOW_PARAMS,
                "sysparm_query": "ip_addressLIKE{observable}",
            },
            headers=_DEF_HEADERS,
        ),
        "ipv6": _ServiceNowParams(
            path="/cmdb_ci_computer",
            params={
                **_SERVICE_NOW_PARAMS,
                "sysparm_query": "ip_addressLIKE{observable}",
            },
            headers=_DEF_HEADERS,
        ),
        "hostname": _ServiceNowParams(
            path="/cmdb_ci_computer",
            params={**_SERVICE_NOW_PARAMS, "sysparm_query": "nameLIKE{observable}"},
            headers=_DEF_HEADERS,
        ),
        "email": _ServiceNowParams(
            path="/sys_user",
            params={**_SERVICE_NOW_PARAMS, "sysparm_query": "email={observable}"},
            headers=_DEF_HEADERS,
        ),
        "user": _ServiceNowParams(
            path="/sys_user",
            params={**_SERVICE_NOW_PARAMS, "sysparm_query": "nameLIKE{observable}"},
            headers=_DEF_HEADERS,
        ),
    }

    _REQUIRED_PARAMS = ["ApiID", "AuthKey", "Instance"]

    def parse_results(self, response: Dict[str, Any]) -> Tuple[bool, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, Any]
            bool = positive or negative hit
            Object with match details

        """
        if self._failed_response(response) or not isinstance(
            response.get("RawResult", {}), dict
        ):
            return False, "Not found."

        if response["ObservableType"] in ["ipv4", "ipv6"]:
            # As Service Now IP field is a CSV-like string and not a list
            # the IP returned might not be an exact match.
            # Therefore, we must validate before returning it
            response["RawResult"]["result"] = [
                result
                for result in response["RawResult"]["result"]
                if response["Observable"]
                in (result.get("ip_address", "").replace(" ", "").split(","))
            ]
            if not response["RawResult"]["result"]:
                return False, "Not found."

        results = response["RawResult"]["result"]

        result_dict = [
            {
                "permalink": f"{response['Reference']}/{result.get('sys_id')}",
                "sys_class_name": result.get("sys_class_name", ""),
                "sys_id": result.get("sys_id", ""),
                "sys_tags": result.get("sys_tags", ""),
                "company": result.get("company", None),
                "location": result.get("location", ""),
                "sys_updated_on": (
                    dt.datetime.strptime(result["sys_updated_on"], "%Y-%m-%d %H:%M:%S")
                    if "sys_updated_on" in result and result["sys_updated_on"]
                    else ""
                ),
                "sys_created_on": (
                    dt.datetime.strptime(result["sys_created_on"], "%Y-%m-%d %H:%M:%S")
                    if "sys_created_on" in result and result["sys_created_on"]
                    else ""
                ),
                **(
                    getattr(self, f"_parse_result_{response['ObservableType']}")(result)
                ),
            }
            for result in results
        ]

        return True, result_dict

    def _parse_result_ipv4(self, result: Dict[str, Any]) -> Dict[str, Any]:
        """
        Return a dictionary of relevant fields for a "Computer" Service Now object.

        Parameters
        ----------
        result : Dict[str, Any]
            Entry returned by Service Now

        Returns
        -------
        Dict[str, Any]

            Subset of releavnt keys parsed

        """
        return {
            "assignment_group": result.get("assignment_group", None),
            "busines_criticality": result.get("busines_criticality", ""),
            "category": result.get("category", None),
            "comments": result.get("comments", None),
            "fqdn": result.get("fqdn", None),
            "hostname": result.get("hostname", None),
            "install_status": result.get("install_status", ""),
            "internet_facing": result.get("internet_facing", False),
            "ip_address": result.get("ip_address", "").split(","),
            "subcategory": result.get("subcategory", None),
            "support_group": result.get("support_group", None),
            "sys_class_name": result.get("sys_class_name", ""),
            "last_discovered": (
                dt.datetime.strptime(result["last_discovered"], "%Y-%m-%d %H:%M:%S")
                if "last_discovered" in result and result["last_discovered"]
                else ""
            ),
            "install_date": (
                dt.datetime.strptime(result["install_date"], "%Y-%m-%d %H:%M:%S")
                if "install_date" in result and result["install_date"]
                else ""
            ),
            "created_on": (
                dt.datetime.strptime(result["created_on"], "%Y-%m-%d %H:%M:%S")
                if "created_on" in result and result["created_on"]
                else ""
            ),
        }

    def _parse_result_user(self, result: Dict[str, Any]) -> Dict[str, Any]:
        """
        Return a dictionary of relevant fields for a "Computer" Service Now object.

        Parameters
        ----------
        result : Dict[str, Any]
            Entry returned by Service Now

        Returns
        -------
        Dict[str, Any]

            Subset of releavnt keys parsed

        """
        return {
            "account": result.get("employee_number", ""),
            "active": result.get("active", "false"),
            "business_criticality": result.get("business_criticality", ""),
            "department": result.get("department", ""),
            "email": result.get("email", ""),
            "failed_attempts": result.get("failed_attempts", 0),
            "first_name": result.get("first_name", ""),
            "introduction": result.get("introduction", ""),
            "last_name": result.get("last_name", ""),
            "manager": result.get("manager", ""),
            "middle_name": result.get("middle_name", ""),
            "name": result.get("name", ""),
            "phone": result.get("phone", "false"),
            "preferred_language": result.get("preferred_language", ""),
            "title": result.get("title", ""),
            "user_name": result.get("user_name", ""),
            "vip": result.get("vip", "false"),
            "last_login": (
                dt.datetime.strptime(result["last_login"], "%Y-%m-%d")
                if "last_login" in result and result["last_login"]
                else ""
            ),
            "last_login_time": (
                dt.datetime.strptime(result["last_login_time"], "%Y-%m-%d %H:%M:%S")
                if "last_login_time" in result and result["last_login_time"]
                else ""
            ),
        }

    # Define aliases
    _parse_result_ipv6 = _parse_result_ipv4
    _parse_result_dns = _parse_result_ipv4
    _parse_result_hostname = _parse_result_ipv4
    _parse_result_email = _parse_result_user



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Search Features."""
from datetime import datetime, timedelta
from uuid import uuid4

import httpx
from azure.common.exceptions import CloudError

from ..._version import VERSION
from .azure_data import get_api_headers
from .sentinel_utils import _build_sent_data

__version__ = VERSION
__author__ = "Pete Bryan"


class SentinelSearchlistsMixin:
    """Mixin class for Sentinel Watchlist feature integrations."""

    def create_search(
        self,
        query: str,
        start: datetime = None,
        end: datetime = None,
        search_name: str = None,
        **kwargs,
    ):
        """
        Create a Search job.

        Parameters
        ----------
        query : str
            The KQL query to run as a job.
        start : datetime, optional
            The start time for the query, by default 90 days ago.
        end : datetime, optional
            The end time for the query, by default now.
        search_name : str, optional
            A name to apply to the search, by default a random GUID is generated.

        Raises
        ------
        CloudError
            If there is an error creating the search job.

        """
        limit = 1000
        if "limit" in kwargs:
            limit = kwargs.pop("limit")
        if "timespan" in kwargs:
            start = kwargs.get("timespan").start  # type: ignore
            end = kwargs.get("timespan").end  # type: ignore
        search_end = end or datetime.now()
        search_start = start or (search_end - timedelta(days=90))
        search_name = search_name or uuid4()  # type: ignore
        search_name = search_name.replace("_", "")  # type: ignore
        search_url = (
            self.sent_urls["search"]  # type: ignore
            + f"/{search_name}_SRCH?api-version=2021-12-01-preview"
        )
        search_items = {
            "searchResults": {
                "query": f"{query}",
                "limit": limit,
                "startSearchTime": f"{search_start.isoformat()}",
                "endSearchTime": f"{search_end.isoformat()}",
            }
        }
        search_body = _build_sent_data(search_items)
        search_create_response = httpx.put(
            search_url,
            headers=get_api_headers(self._token),  # type: ignore
            json=search_body,
            timeout=60,
        )
        if search_create_response.status_code != 202:
            raise CloudError(response=search_create_response)
        print(f"Search job created with for {search_name}_SRCH.")

    def check_search_status(self, search_name: str) -> bool:
        """
        Check the status of a search job.

        Parameters
        ----------
        search_name : str
            The name of the search job to check.

        Returns
        -------
        bool
            Returns True if search is ready.

        Raises
        ------
        CloudError
            If error in checking the search job status.

        """
        search_name = search_name.strip("_SRCH")
        search_url = (
            self.sent_urls["search"]  # type: ignore
            + f"/{search_name}_SRCH?api-version=2021-12-01-preview"
        )
        search_check_response = httpx.get(
            search_url, headers=get_api_headers(self._token)  # type: ignore
        )
        if search_check_response.status_code != 200:
            raise CloudError(response=search_check_response)

        check_result = search_check_response.json()["properties"]["provisioningState"]
        print(f"{search_name}_SRCH status is '{check_result}'")
        if check_result == "Succeeded":
            return True
        return False

    def delete_search(self, search_name: str):
        """
        Delete a search result.

        Parameters
        ----------
        search_name: str
            The name of the search to delete.

        Raises
        ------
        CloudError
            If an error occurs when attempting to delete the search

        """
        search_name = search_name.strip("_SRCH")
        search_url = (
            self.sent_urls["search"]  # type: ignore
            + f"/{search_name}_SRCH?api-version=2021-12-01-preview"
        )
        search_delete_response = httpx.delete(
            search_url, headers=get_api_headers(self._token)  # type: ignore
        )
        if search_delete_response.status_code != 202:
            raise CloudError(response=search_delete_response)
        print(f"{search_name}_SRCH set for deletion.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Packaging utility functions."""
import importlib
import os
import re
import subprocess  # nosec
import sys
import warnings
from pathlib import Path
from platform import python_version
from typing import Dict, List, Optional, Tuple, Union

import pkg_resources
from IPython import get_ipython
from IPython.display import HTML, display
from tqdm.auto import tqdm
from tqdm.notebook import tqdm as tqdm_notebook

from ... import _version
from ..._version import VERSION
from .ipython import is_ipython
from .types import export

__version__ = VERSION
__author__ = "Ian Hellen"


@export
def resolve_pkg_path(part_path: str):
    """
    Resolve a path relative to the package.

    Parameters
    ----------
    part_path : str
        Absolute or relative path to resolve.

    """
    if Path(part_path).is_absolute():
        return part_path

    resolved_path = str(Path(_version.__file__).resolve().parent.joinpath(part_path))
    if Path(resolved_path).exists():
        return resolved_path

    searched_paths = list(
        {
            str(path)
            for path in Path(_version.__file__)
            .resolve()
            .parent.glob(str(Path("**").joinpath(part_path)))
            if ".mypy_cache" not in str(path)
        }
    )

    if not searched_paths or len(searched_paths) > 1:
        warnings.warn(f"No path or ambiguous match for {part_path} not found")
        return None
    return str(searched_paths[0])


@export
def check_py_version(min_ver: Tuple = (3, 6)):
    """
    Check that the current python version is not less than `min_ver`.

    Parameters
    ----------
    min_ver : Tuple, optional
        Minimum required version, by default (3,6)

    """
    if isinstance(min_ver, (float, str)):
        min_ver_list = str(min_ver).split(".")
        min_ver = (int(min_ver_list[0]), int(min_ver_list[1]))
    if sys.version_info < min_ver:
        print("Check the Kernel->Change Kernel menu and ensure that Python 3.6")
        print("or later is selected as the active kernel.")
        raise SystemExit(f"Python {min_ver[0]}.{min_ver[1]} or later is required.\n")


# pylint: disable=not-an-iterable, too-many-branches
@export  # noqa: MC0001
def check_and_install_missing_packages(  # noqa: MC0001
    required_packages: List[str],
    force_notebook: bool = False,
    user: bool = False,
    upgrade: bool = False,
) -> bool:
    """
    Check and install missing packages from provided list of packages.

    Parameters
    ----------
    required_packages : List[str]
        List of packages to check and install in a current environment
        Note you can add package version constraints by appending them to
        the package name, e.g. `pandas>=1.01`
    force_notebook : bool, optional
        Boolean value to force notebook version of progress bar,
        by default False (autodetect)
    user : bool, optional
        Boolean value to toggle user flag while installing pip packages,
        by default False
    upgrade: bool, option
        If true supply `--upgrade` flag to pip to install the latest
        version (applies to all package in `required_packages`)

    Returns
    -------
    bool :
        True if successful, else False

    """
    missing_packages = []
    if isinstance(required_packages, str):
        if "," in required_packages:
            required_packages = [
                req.strip() for req in required_packages.split(",") if req.strip()
            ]
        else:
            required_packages = [required_packages]
    # Check package requirements against installed set
    for req in required_packages:
        pkg_req = pkg_resources.Requirement.parse(req)
        try:
            found_pkg = pkg_resources.working_set.find(pkg_req)
        except pkg_resources.VersionConflict:
            found_pkg = None
        if found_pkg is None:
            missing_packages.append(req)

    if not missing_packages:
        print("All packages are already installed")
        return True

    print("Missing packages to be installed: ", *missing_packages, sep=" ")
    if is_ipython() or force_notebook:
        pkgbar = tqdm_notebook(missing_packages, desc="Installing...", unit="bytes")
    else:
        pkgbar = tqdm(missing_packages, desc="Installing...", unit="bytes")

    pkg_command = ["install"] if is_ipython() else ["python", "-m", "pip", "install"]
    if user:
        pkg_command.append("--user")
    if upgrade:
        pkg_command.append("--upgrade")
    pkg_success = True
    for package in pkgbar:
        if is_ipython():
            get_ipython().run_line_magic("pip", " ".join(pkg_command + [package]))
        else:
            try:
                subprocess.run(  # nosec
                    pkg_command + [package],
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            except subprocess.CalledProcessError as proc_err:
                print(f"An Error has occurred while installing {package}.")
                print(f"Output: {proc_err.stdout}")
                print(f"Errs: {proc_err.stderr}")
                pkg_success = False
        print(f"{package} installed.")

    return pkg_success


# pylint: enable=not-an-iterable, too-many-branches


def _get_mp_ua():
    """Build a MSTICPy specific User Agent string."""
    return f"MSTICPy{VERSION}-Python{python_version()}-{sys.platform}"


# User Agent string for MSTICPy
_MSTICPY_USER_AGENT = _get_mp_ua()


@export
def mp_ua_header() -> Dict[str, str]:
    """Return headers dict for MSTICPy User Agent."""
    return {"UserAgent": _get_mp_ua()}


@export
def search_for_file(
    pattern: str, paths: List[Union[str, Path]] = None
) -> Optional[str]:
    """Search `paths` for file `pattern`."""
    paths = paths or [".", ".."]
    for start_path in paths:
        found_files = list(Path(start_path).glob(pattern)) if start_path else []
        if found_files:
            return str(found_files[0])
    return None


@export
def search_module(pattern: str) -> Dict[str, str]:
    """
    Return MSTICPy modules that match `pattern`.

    Parameters
    ----------
    pattern : str
        Substring or regular expression

    Returns
    -------
    Dict[str, str]
        Dict of module name and help URI of matching modules

    """
    pkg_path = Path(resolve_pkg_path(""))
    deprec_pattern = "Deprecated - .* has moved"
    html_tmplt = "https://msticpy.readthedocs.io/en/latest/api/{mod_name}.html"

    def _format_module(path, root):
        stem = str(path).replace(str(root), "")
        stem = stem[1:] if stem[0] in ("\\", "/") else stem
        return stem.replace(".py", "").replace("/", ".").replace("\\", ".")

    module_list = {
        _format_module(file, pkg_path.parent)
        for file in Path(pkg_path).rglob("*.py")
        if not str(file).endswith("__init__.py")
        and re.search(pattern, str(file))
        and not re.search(deprec_pattern, file.read_text())
    }
    return {module: html_tmplt.format(mod_name=module) for module in module_list}


@export
def search_name(pattern: str) -> None:
    """
    Display matching modules as list or HTML table.

    Parameters
    ----------
    pattern : str
        Substring or regular expression

    """
    mod_uris = search_module(pattern)
    if not is_ipython(notebook=True):
        print("\n".join(f"{mod}:\t{uri}" for mod, uri in mod_uris.items()))
        return

    table_template = """
    <style>
        .table_mod {{border-collapse: collapse; width: 50%;}}
        .cell_mod {{border: 1px solid #ddd !important;
            text-align: left !important; padding: 15px !important;}}
    </style>
    <h4>Modules matching '{pattern}'</h4>
    <table class='table_mod'>
    <tr class='cell_mod'><th>Module</th><th>Help</th></tr>
    {rows}
    </table>
    """
    rows = [
        (
            f"<tr class='cell_mod'><td>{mod}</td><td>"
            f"<a href={uri} target='_blank'>{mod}</a></td></tr>"
        )
        for mod, uri in mod_uris.items()
    ]
    display(HTML(table_template.format(rows="\n".join(rows), pattern=pattern)))


_U_TEST_ENV = "PYTEST_CURRENT_TEST"


@export
def unit_testing() -> bool:
    """
    Return True if in unit testing.

    Returns
    -------
    bool
        True if in unit testing

    """
    return _U_TEST_ENV in os.environ


# pylint: disable=invalid-name
@export
def set_unit_testing(on: bool = True):
    """
    Set flag env var to indicated that code is being unit-tested.

    Parameters
    ----------
    on : bool, optional
        Turn unit testing flag on or off, by default True

    """
    if on:
        os.environ[_U_TEST_ENV] = "True"
    else:
        os.environ.pop(_U_TEST_ENV, None)


def init_getattr(module_name: str, dynamic_imports: Dict[str, str], attrib: str):
    """Import and return dynamic attribute."""
    if attrib in dynamic_imports:
        module = importlib.import_module(dynamic_imports[attrib])
        return getattr(module, attrib)
    raise AttributeError(f"{module_name} has no attribute {attrib}")


def init_dir(static_attribs: List[str], dynamic_imports: Dict[str, str]):
    """Return list of available attributes."""
    return sorted(set(static_attribs + list(dynamic_imports)))


def delayed_import(module: str, attrib: str, call: bool = False):
    """Import attribute from module on demand."""
    attribute = None

    def import_item(*args, **kwargs):
        """Return the attribute, importing module if needed."""
        nonlocal attribute
        if attribute is None:
            imp_module = importlib.import_module(module)
            attribute = getattr(imp_module, attrib)
        return (
            attribute(*args, **kwargs) if (call and callable(attribute)) else attribute
        )

    return import_item



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Kql TI Provider base.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""

import abc
import contextlib
import warnings
from collections import defaultdict
from functools import lru_cache
from typing import Any, Callable, DefaultDict, Dict, Iterable, List, Set, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ...common.exceptions import MsticpyConfigError
from ...common.utility import export
from ...common.wsconfig import WorkspaceConfig
from ...data.core.data_providers import QueryProvider
from ..lookup_result import LookupStatus
from ..provider_base import generate_items
from .ti_provider_base import ResultSeverity, TIProvider

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class KqlTIProvider(TIProvider):
    """KQL TI provider base class."""

    _QUERIES: Dict[str, tuple] = {}

    _CONNECT_STR = (
        "loganalytics://code().tenant('{TENANT_ID}').workspace('{WORKSPACE_ID}')"
    )

    _REQUIRED_TABLES: List[str] = []

    def __init__(self, **kwargs):
        """Initialize a new instance of the class."""
        super().__init__(**kwargs)

        if "query_provider" in kwargs and isinstance(
            kwargs["query_provider"], QueryProvider
        ):
            self._query_provider = kwargs.pop("query_provider")
            self._connect_str = kwargs.pop("connect_str", WorkspaceConfig())
        else:
            self._query_provider, self._connect_str = self._create_query_provider(
                **kwargs
            )

        if not self._query_provider:
            raise MsticpyConfigError("Query provider for KQL could not be created.")

    @property
    def _connected(self):
        return self._query_provider.connected

    @lru_cache(maxsize=256)
    def lookup_ioc(  # type: ignore
        self, ioc: str, ioc_type: str = None, query_type: str = None, **kwargs
    ) -> pd.DataFrame:
        """
        Lookup from a value.

        Parameters
        ----------
        ioc : str
            item to lookup
        ioc_type : str, optional
            The Type of the item to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """
        return self.lookup_iocs(
            data={ioc: ioc_type},
            query_type=query_type,
            **kwargs,
        )

    # pylint: disable=too-many-locals
    def lookup_iocs(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup collection of IoC observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        if not self._connected:
            self._connect()
        if any(
            table not in self._query_provider.schema for table in self._REQUIRED_TABLES
        ):
            return pd.DataFrame()

        # We need to partition the IoC types to invoke separate queries
        ioc_groups: DefaultDict[str, Set[str]] = defaultdict(set)
        for ioc, ioc_type in generate_items(data, ioc_col, ioc_type_col):
            if not ioc:
                continue
            result = self._check_ioc_type(ioc, ioc_type, query_type)

            if result["Status"] != LookupStatus.NOT_SUPPORTED.value:
                ioc_groups[result["IocType"]].add(result["Ioc"])

        all_results = []
        for ioc_type, obs_set in ioc_groups.items():
            kwargs.pop("ioc_type", None)  # we want to use locally-determine type
            query_obj = None
            with contextlib.suppress(LookupError):
                query_obj, query_params = self._get_query_and_params(
                    ioc=list(obs_set),
                    ioc_type=ioc_type,
                    query_type=query_type,
                    **kwargs,
                )
            if not query_obj:
                warnings.warn(f"Could not find query for {ioc_type}, {query_type}")
                continue

            # run the query
            data_result = query_obj(**query_params)

            src_ioc_frame = pd.DataFrame(obs_set, columns=["Ioc"])
            src_ioc_frame["IocType"] = ioc_type
            src_ioc_frame["QuerySubtype"] = query_type
            src_ioc_frame["Reference"] = query_obj("print_query", **query_params)

            lookup_status = self._check_result_status(data_result)
            # If no results, add the empty dataframe to the combined results
            # and continue
            if lookup_status in {LookupStatus.QUERY_FAILED, LookupStatus.NO_DATA}:
                self._add_failure_status(src_ioc_frame, lookup_status)
                all_results.append(src_ioc_frame)
                continue

            # copy the DF to avoid pandas warnings about changing DF view
            data_result = data_result.copy()
            # Create our results columns
            data_result["Result"] = True
            data_result["Status"] = LookupStatus.OK.value
            data_result["Severity"] = self._get_severity(data_result)
            data_result["Details"] = self._get_detail_summary(data_result)
            data_result["RawResult"] = data_result.apply(lambda x: x.to_dict(), axis=1)

            combined_results_df = self._combine_results(
                input_df=src_ioc_frame, results_df=data_result, reslt_ioc_key="IoC"
            )
            all_results.append(combined_results_df)

        if all_results:
            return pd.concat(all_results, ignore_index=True, sort=False, axis=0)
        return pd.DataFrame()

    @staticmethod
    def _add_failure_status(src_ioc_frame: pd.DataFrame, lookup_status: LookupStatus):
        """Add status info, if query produced no results."""
        src_ioc_frame["Result"] = False
        src_ioc_frame["Details"] = (
            "Query failure"
            if lookup_status == LookupStatus.QUERY_FAILED
            else "Not found"
        )
        src_ioc_frame["Status"] = lookup_status.value
        src_ioc_frame["Severity"] = ResultSeverity.information.name

    @staticmethod
    def _check_result_status(data_result):
        """Check the return value from the query."""
        if isinstance(data_result, pd.DataFrame):
            return LookupStatus.NO_DATA if data_result.empty else LookupStatus.OK
        if (
            hasattr(data_result, "completion_query_info")
            and data_result.completion_query_info["StatusCode"] == 0
            and data_result.records_count == 0
        ):
            print("No results return from data provider.")
            return LookupStatus.NO_DATA
        if data_result and hasattr(data_result, "completion_query_info"):
            print(
                "No results returned from data provider. "
                + str(data_result.completion_query_info)
            )
        else:
            print(f"Unknown response from provider: {str(data_result)}")
        return LookupStatus.QUERY_FAILED

    @abc.abstractmethod
    def parse_results(self, response: Dict) -> Tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """

    @staticmethod
    @abc.abstractmethod
    def _get_detail_summary(data_result: pd.DataFrame) -> pd.Series:
        pass

    @staticmethod
    @abc.abstractmethod
    def _get_severity(data_result: pd.DataFrame) -> pd.Series:
        pass

    def _create_query_provider(self, **kwargs):
        workspace_id = None
        tenant_id = None
        workspace_id = self._get_spelled_variants("workspaceid", **kwargs)
        tenant_id = self._get_spelled_variants("tenantid", **kwargs)

        if not workspace_id or not tenant_id:
            # If there are no TI-Provider specific kwargs
            # WorkspaceConfig should be able to get these global values
            # If a config file or a workspace name is passed we'll use
            # those in case there are multiple workspaces set globally.
            config_file = kwargs.get("config_file")
            workspace = kwargs.get("workspace")
            ws_config = WorkspaceConfig(config_file=config_file, workspace=workspace)
            workspace_id = ws_config["workspace_id"]
            tenant_id = ws_config["tenant_id"]
        # Either the format or connect() call will fail if these values are
        # not set or invalid.
        connect_str = self._CONNECT_STR.format(
            TENANT_ID=tenant_id, WORKSPACE_ID=workspace_id
        )
        query_provider = QueryProvider("LogAnalytics")
        return query_provider, connect_str

    def _connect(self):
        """Connect to query provider."""
        print("MS Sentinel TI query provider needs authenticated connection.")
        self._query_provider.connect(self._connect_str)

    @staticmethod
    def _get_spelled_variants(name: str, **kwargs) -> Any:
        """Return value with matching variant spelling key."""
        variant_dict = {
            "workspaceid": ["workspace_id", "workspaceid"],
            "tenantid": ["tenant_id", "tenantid"],
        }
        variants = variant_dict.get(name, [name.casefold()])
        return next(
            (val for key, val in kwargs.items() if key.casefold() in variants), None
        )

    def _get_query_and_params(
        self,
        ioc: Union[str, List[str]],
        ioc_type: str,
        query_type: str = None,
        **kwargs,
    ) -> Tuple[Callable, Dict[str, Any]]:
        ioc_key = f"{ioc_type}-{query_type}" if query_type else ioc_type
        query_def = self._QUERIES.get(ioc_key, None)
        if not query_def:
            raise LookupError(f"Provider does not support IoC type {ioc_key}.")

        query_name = query_def[0]
        query_def_params = query_def[1]
        if "ioc" not in query_def_params:
            raise ValueError(
                f"No parameter name defined for observable for {ioc_type}. "
                + f"Referenced query: {query_name}"
            )

        query_params = {query_def_params["ioc"]: ioc}
        if "start" in kwargs:
            query_params["start"] = kwargs["start"]
        if "end" in kwargs:
            query_params["end"] = kwargs["end"]

        query_obj = getattr(self._query_provider, query_name, None)
        if not query_obj:
            raise ValueError(
                f"No query object name for {query_name} found in provider."
            )

        return query_obj, query_params

    @staticmethod
    def _series_to_list(series: pd.Series) -> List[Any]:
        return list(series.dropna().unique())

    @staticmethod
    def _combine_results(
        input_df: pd.DataFrame, results_df: pd.DataFrame, reslt_ioc_key: str
    ) -> pd.DataFrame:
        # Clean out unwanted columns from the results and merge with
        # the original IoCList
        # If we have results, we need to create our summary columns
        # merge the results with our original IoC set
        # and drop all of the columns that we are not interested in
        columns_to_drop = set(results_df.columns.to_list())
        colums_to_keep = {
            reslt_ioc_key,
            "Result",
            "Status",
            "Severity",
            "Details",
            "RawResult",
        }
        columns_to_drop = columns_to_drop - colums_to_keep

        cleaned_results_df = results_df.copy().drop(columns=columns_to_drop)  # type: ignore
        combined_df = input_df.copy()
        combined_df["IoCKey"] = input_df["Ioc"].str.lower()
        cleaned_results_df = cleaned_results_df.rename(
            columns={reslt_ioc_key: "IoCKey"}
        )
        combined_df = combined_df.merge(
            right=cleaned_results_df, how="left", on="IoCKey"
        ).drop(columns="IoCKey")
        # Fill in any NaN values from the merge
        combined_df["Result"] = combined_df["Result"].fillna(False)
        combined_df["Details"] = combined_df["Details"].fillna("Not found.")
        combined_df["Status"] = combined_df["Status"].fillna(LookupStatus.OK.value)
        combined_df["Severity"] = combined_df["Severity"].fillna(
            ResultSeverity.information.value
        )
        return combined_df



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data provider loader."""
import logging
from functools import partial
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ...common.pkg_config import get_config
from ...common.utility import export, valid_pyname
from ...nbwidgets.query_time import QueryTime
from .. import drivers
from ..drivers.driver_base import DriverBase, DriverProps
from .param_extractor import extract_query_params
from .query_container import QueryContainer
from .query_defns import DataEnvironment
from .query_provider_connections_mixin import QueryProviderConnectionsMixin
from .query_provider_utils_mixin import QueryProviderUtilsMixin
from .query_store import QueryStore

__version__ = VERSION
__author__ = "Ian Hellen"

_HELP_FLAGS = ("help", "?")
_DEBUG_FLAGS = ("print", "debug_query", "print_query")
_COMPATIBLE_DRIVER_MAPPINGS = {
    "mssentinel": ["m365d"],
    "mde": ["m365d"],
    "mssentinel_new": ["mssentinel", "m365d"],
    "kusto_new": ["kusto"],
}

logger = logging.getLogger(__name__)


# These are mixin classes that do not have an __init__ method
# pylint: disable=super-init-not-called
@export
class QueryProvider(QueryProviderConnectionsMixin, QueryProviderUtilsMixin):
    """
    Container for query store and query execution provider.

    Instances of this class hold the query set and execution
    methods for a specific data environment.

    """

    def __init__(  # noqa: MC0001
        self,
        data_environment: Union[str, DataEnvironment],
        driver: Optional[DriverBase] = None,
        query_paths: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Query provider interface to queries.

        Parameters
        ----------
        data_environment : Union[str, DataEnvironment]
            Name or Enum of environment for the QueryProvider
        driver : DriverBase, optional
            Override the builtin driver (query execution class)
            and use your own driver (must inherit from
            `DriverBase`)
        query_paths : List[str]
            Additional paths to look for query definitions.
        kwargs :
            Other arguments are passed to the data provider driver.

        Notes
        -----
        Additional keyword arguments are passed to the data provider driver.
        The driver may support additional keyword arguments, use
        the QueryProvider.driver_help() method to get a list of these
        parameters.

        See Also
        --------
        DataProviderBase : base class for data query providers.

        """
        # import at runtime to prevent circular import
        # pylint: disable=import-outside-toplevel, cyclic-import
        from ...init.pivot_init.pivot_data_queries import add_data_queries_to_entities

        # pylint: enable=import-outside-toplevel
        setattr(self.__class__, "_add_pivots", add_data_queries_to_entities)

        data_environment, self.environment_name = self._check_environment(
            data_environment
        )

        self._driver_kwargs = kwargs.copy()
        if driver is None:
            self.driver_class = drivers.import_driver(data_environment)
            if issubclass(self.driver_class, DriverBase):
                driver = self.driver_class(data_environment=data_environment, **kwargs)
            else:
                raise LookupError(
                    "Could not find suitable data provider for", f" {data_environment}"
                )
        else:
            self.driver_class = driver.__class__
        # allow the driver to override the data environment used for selecting queries
        self.environment_name = (
            driver.get_driver_property(DriverProps.EFFECTIVE_ENV)
            or self.environment_name
        )
        logger.info("Using data environment %s", self.environment_name)
        logger.info("Driver class: %s", self.driver_class.__name__)

        self._additional_connections: Dict[str, DriverBase] = {}
        self._query_provider = driver
        # replace the connect method docstring with that from
        # the driver's connect method
        self.__class__.connect.__doc__ = driver.connect.__doc__
        self.all_queries = QueryContainer()

        # Add any query files
        data_env_queries: Dict[str, QueryStore] = {}
        self._query_paths = query_paths
        if driver.use_query_paths:
            logger.info("Using query paths %s", query_paths)
            data_env_queries.update(
                self._read_queries_from_paths(query_paths=query_paths)
            )
        self.query_store = data_env_queries.get(
            self.environment_name, QueryStore(self.environment_name)
        )
        logger.info("Adding query functions to provider")
        self._add_query_functions()
        self._query_time = QueryTime(units="day")
        logger.info("Initialization complete.")

    def _check_environment(
        self, data_environment
    ) -> Tuple[Union[str, DataEnvironment], str]:
        """Check environment against known names."""
        if isinstance(data_environment, str):
            data_env = DataEnvironment.parse(data_environment)
            if data_env != DataEnvironment.Unknown:
                data_environment = data_env
                environment_name = data_environment.name
            elif data_environment in drivers.CUSTOM_PROVIDERS:
                environment_name = data_environment
            else:
                raise TypeError(f"Unknown data environment {data_environment}")
        elif isinstance(data_environment, DataEnvironment):
            environment_name = data_environment.name
        else:
            raise TypeError(
                f"Unknown data environment type {data_environment} ({type(data_environment)})"
            )
        return data_environment, environment_name

    def __getattr__(self, name):
        """Return the value of the named property 'name'."""
        if "." in name:
            parent_name, child_name = name.split(".", maxsplit=1)
            parent = getattr(self, parent_name, None)
            if parent:
                return getattr(parent, child_name)
        raise AttributeError(f"{name} is not a valid attribute.")

    @property
    def environment(self) -> str:
        """Return the environment name."""
        return self.environment_name

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connection string for the data source

        """
        logger.info("Calling connect on driver")
        self._query_provider.connect(connection_str=connection_str, **kwargs)

        # If the driver has any attributes to expose via the provider
        # add those here.
        for attr_name, attr in self._query_provider.public_attribs.items():
            setattr(self, attr_name, attr)

        refresh_query_funcs = False
        # if the driver supports dynamic filtering of queries,
        # filter the query store based on connect-time parameters
        if self._query_provider.filter_queries_on_connect:
            self.query_store.apply_query_filter(self._query_provider.query_usable)
            refresh_query_funcs = True
        # Add any built-in or dynamically retrieved queries from driver
        if self._query_provider.has_driver_queries:
            logger.info("Adding driver queries to provider")
            driver_queries = self._query_provider.driver_queries
            self._add_driver_queries(queries=driver_queries)
            refresh_query_funcs = True

        if refresh_query_funcs:
            self._add_query_functions()

        # Since we're now connected, add Pivot functions
        logger.info("Adding query pivot functions")
        self._add_pivots(lambda: self._query_time.timespan)

    def exec_query(self, query: str, **kwargs) -> Union[pd.DataFrame, Any]:
        """
        Execute simple query string.

        Parameters
        ----------
        query : str
            [description]
        use_connections : Union[str, List[str]]

        Other Parameters
        ----------------
        query_options : Dict[str, Any]
            Additional options passed to query driver.
        kwargs : Dict[str, Any]
            Additional options passed to query driver.

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results - a DataFrame if successful
            or a KqlResult if unsuccessful.

        """
        query_options = kwargs.pop("query_options", {}) or kwargs
        query_source = kwargs.pop("query_source", None)

        logger.info("Executing query '%s...'", query[:40])
        logger.debug("Full query: %s", query)
        logger.debug("Query options: %s", query_options)
        if not self._additional_connections:
            return self._query_provider.query(
                query, query_source=query_source, **query_options
            )
        return self._exec_additional_connections(query, **kwargs)

    @property
    def query_time(self):
        """Return the default QueryTime control for queries."""
        return self._query_time

    def _execute_query(self, *args, **kwargs) -> Union[pd.DataFrame, Any]:
        if not self._query_provider.loaded:
            raise ValueError("Provider is not loaded.")
        if (
            not self._query_provider.connected
            and not _help_flag(*args)
            and not _debug_flag(*args, **kwargs)
        ):
            raise ValueError(
                "No connection to a data source.",
                "Please call connect(connection_str) and retry.",
            )
        query_name = kwargs.pop("query_name")
        family = kwargs.pop("query_path")

        query_source = self.query_store.get_query(
            query_path=family, query_name=query_name
        )
        if _help_flag(*args):
            query_source.help()
            return None

        params, missing = extract_query_params(query_source, *args, **kwargs)
        logger.debug("Template query: %s", query_source.query)
        logger.info("Parameters for query: %s", params)
        query_options = {
            "default_time_params": self._check_for_time_params(params, missing)
        }
        if missing:
            query_source.help()
            raise ValueError(f"No values found for these parameters: {missing}")

        split_by = kwargs.pop("split_query_by", kwargs.pop("split_by", None))
        if split_by:
            logger.info("Split query selected - interval - %s", split_by)
            split_result = self._exec_split_query(
                split_by=split_by,
                query_source=query_source,
                query_params=params,
                debug=_debug_flag(*args, **kwargs),
                args=args,
                **kwargs,
            )
            if split_result is not None:
                return split_result
            # if split queries could not be created, fall back to default
        query_str = query_source.create_query(
            formatters=self._query_provider.formatters, **params
        )
        # This looks for any of the "print query" debug args in args or kwargs
        if _debug_flag(*args, **kwargs):
            return query_str

        # Handle any query options passed and run the query
        query_options.update(self._get_query_options(params, kwargs))
        logger.info(
            "Running query '%s...' with params: %s", query_str[:40], query_options
        )
        return self.exec_query(query_str, query_source=query_source, **query_options)

    def _check_for_time_params(self, params, missing) -> bool:
        """Fall back on builtin query time if no time parameters were supplied."""
        defaults_added = False
        if "start" in missing:
            missing.remove("start")
            params["start"] = self._query_time.start
            defaults_added = True
        if "end" in missing:
            missing.remove("end")
            params["end"] = self._query_time.end
            defaults_added = True
        return defaults_added

    def _get_query_folder_for_env(self, root_path: str, environment: str) -> List[Path]:
        """Return query folder for current environment."""
        data_envs = [environment]
        if environment.casefold() in _COMPATIBLE_DRIVER_MAPPINGS:
            data_envs += _COMPATIBLE_DRIVER_MAPPINGS[environment.casefold()]
        return [
            _resolve_package_path(root_path).joinpath(data_env.casefold())
            for data_env in data_envs
        ]

    def _read_queries_from_paths(self, query_paths) -> Dict[str, QueryStore]:
        """Fetch queries from YAML files in specified paths."""
        settings: Dict[str, Any] = get_config("QueryDefinitions", {})
        all_query_paths: List[Union[Path, str]] = []
        for def_qry_path in settings.get("Default"):  # type: ignore
            # only read queries from environment folder
            builtin_qry_paths = self._get_query_folder_for_env(
                def_qry_path, self.environment_name
            )
            all_query_paths.extend(
                str(qry_path) for qry_path in builtin_qry_paths if qry_path.is_dir()
            )

        if settings.get("Custom") is not None:
            for custom_path in settings.get("Custom"):  # type: ignore
                custom_qry_path = _resolve_path(custom_path)
                if custom_qry_path:
                    all_query_paths.append(custom_qry_path)
        if query_paths:
            for param_path in query_paths:
                param_qry_path = _resolve_path(param_path)
                if param_qry_path:
                    all_query_paths.append(param_qry_path)
        if all_query_paths:
            logger.info("Reading queries from %s", all_query_paths)
            return QueryStore.import_files(
                source_path=all_query_paths,
                recursive=True,
                driver_query_filter=self._query_provider.query_attach_spec,
            )
        # if no queries - just return an empty store
        return {self.environment_name: QueryStore(self.environment_name)}

    def _add_query_functions(self):
        """Add queries to the module as callable methods."""
        for qual_query_name in self.list_queries():
            query_path = qual_query_name.split(".")
            query_name = query_path[-1]
            current_node = self
            for container_name in query_path[:-1]:
                container_name = valid_pyname(container_name)
                if hasattr(current_node, container_name):
                    current_node = getattr(current_node, container_name)
                else:
                    new_node = QueryContainer()
                    setattr(current_node, container_name, new_node)
                    current_node = new_node

            query_cont_name = ".".join(query_path[:-1])

            # Create the partial function
            query_func = partial(
                self._execute_query, query_path=query_cont_name, query_name=query_name
            )
            query_func.__doc__ = self.query_store.get_query(
                query_path=query_cont_name, query_name=query_name
            ).create_doc_string()

            query_name = valid_pyname(query_name)
            setattr(current_node, query_name, query_func)
            setattr(self.all_queries, query_name, query_func)

    def _add_driver_queries(self, queries: Iterable[Dict[str, str]]):
        """Add driver queries to the query store."""
        for query in queries:
            self.query_store.add_query(
                name=query["name"],
                query=query["query"],
                query_paths=query.get(
                    "query_paths", query.get("query_container", "default")
                ),
                description=query["description"],
            )
        # For now, just add all of the functions again (with any connect-time acquired
        # queries) - we could be more efficient than this but unless there are 1000s of
        # queries it should not be noticeable.
        self._add_query_functions()

    @staticmethod
    def _get_query_options(
        params: Dict[str, Any], kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        # sourcery skip: inline-immediately-returned-variable, use-or-for-fallback
        """Return any kwargs not already in params."""
        query_options = kwargs.pop("query_options", {})
        if not query_options:
            # Any kwargs left over we send to the query provider driver
            query_options = {
                key: val for key, val in kwargs.items() if key not in params
            }
        query_options["time_span"] = {
            "start": params.get("start"),
            "end": params.get("end"),
        }
        return query_options


def _resolve_package_path(config_path: str) -> Path:
    """Resolve path relative to current package."""
    return (
        Path(config_path)
        if Path(config_path).is_absolute()
        else Path(__file__).resolve().parent.parent.joinpath(config_path)
    )


def _resolve_path(config_path: str) -> Optional[str]:
    """Resolve path."""
    if not Path(config_path).is_absolute():
        config_path = str(Path(config_path).expanduser().resolve())
    if not Path(config_path).is_dir():
        print(f"Warning: Custom query definitions path {config_path} not found")
        return None
    return config_path


def _help_flag(*args) -> bool:
    """Return True if help parameter passed."""
    return any(help_flag for help_flag in _HELP_FLAGS if help_flag in args)


def _debug_flag(*args, **kwargs) -> bool:
    """Return True if debug/print args passed."""
    return any(db_arg for db_arg in _DEBUG_FLAGS if db_arg in args) or any(
        db_arg for db_arg in _DEBUG_FLAGS if kwargs.get(db_arg, False)
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Analytics Features."""
from datetime import datetime
from typing import Optional

import httpx
import pandas as pd
from azure.common.exceptions import CloudError

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import _azs_api_result_to_df, _build_sent_data, get_http_timeout

__version__ = VERSION
__author__ = "Pete Bryan"


_INDICATOR_ITEMS = {
    "valid_to": "validUntil",
    "description": "description",
    "threat_types": "threatTypes",
    "name": "displayName",
    "confidence": "confidence",
    "revoked": "revoked",
    "source": "source",
}

_IOC_TYPE_MAPPING = {
    "dns": "domain-name",
    "url": "url",
    "ipv4": "ipv4-addr",
    "ipv6": "ipv6-addr",
    "md5_hash": "MD5",
    "sha1_hash": "SHA-1",
    "sha256_hash": "SHA-256",
}


class SentinelTIMixin:
    """Mixin class for Sentinel Hunting feature integrations."""

    def get_all_indicators(
        self,
        limit: Optional[int] = None,
        orderby: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Return all TI indicators in a Microsoft Sentinel workspace.

        Parameters
        ----------
        limit: int, optional
            If set returns top n results
        orderby : Optional[str], optional
            Order results by a specific column

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        appendix = "/indicators?"
        if limit:
            appendix += f"&$top={limit}"
        if orderby:
            appendix += f"&$orderby={orderby}"
        return self._list_items(  # type: ignore
            item_type="ti", api_version="2021-10-01", appendix=appendix
        )  # type: ignore

    def get_ti_metrics(self) -> pd.DataFrame:
        """
        Return metrics about TI indicators in a Microsoft Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        return self._list_items(  # type: ignore
            item_type="ti", api_version="2021-10-01", appendix="/metrics"
        )  # type: ignore

    def create_indicator(
        self,
        indicator: str,
        ioc_type: str,
        name: str = "TI Indicator",
        confidence: int = 0,
        silent: bool = False,
        **kwargs,
    ) -> str:
        """
        Create a new indicator within the Microsoft Sentinel workspace.

        Parameters
        ----------
        indicator : str
            The indicator to create - i.e. IP address, domain name etc.
        ioc_type : str
            The type of indicator to create - can be:
            "dns", "url", "ipv4", "ipv6", "md5_hash", "sha1_hash", "sha256_hash"
        name : str, optional
            A common name to give to the indicator default is 'TI Indicator'
        confidence : int, optional
            A score between 0-100 of the confidence in the indicator, defualt is 0
        silent : bool, optional
            If True no output is displayed, defaults to False
        description : str, optional
            An description of the indicator
        labels : list, optional
            A list of string object labels to associate with the indicator
        kill_chain_phases : list, optional
            A list of string objects relating to the kill chain phases
            an indicator is assocaited with
        threat_types : list, optional
            A list of threat types associated with the indicator (list of string objects)
        external_references : list, optional
            A list of URLs that provide an external reference for the indicator
        valid_from : datetime, optional
            A datetime from which the indicator is valid from, defaults to now
        valid_to : datetime, optional
            A datetime to which the indicator is valid until

        Returns
        -------
        The ID of the created indicator

        Raises
        ------
        MsticpyUserError
            If invalid ioc_type or confidence value provided
        CloudError
            If API call fails

        """
        self.check_connected()  # type: ignore
        ti_url = self.sent_urls["ti"] + "/createIndicator"  # type: ignore
        params = {"api-version": "2021-10-01"}
        if ioc_type not in _IOC_TYPE_MAPPING:
            raise MsticpyUserError(
                """ioc_type must be one of -
                 'dns', 'url', 'ipv4', 'ipv6', 'md5_hash', 'sha1_hash', 'sha256_hash'"""
            )
        normalized_ioc_type = _IOC_TYPE_MAPPING[ioc_type]
        pattern_type = normalized_ioc_type
        value = "value"
        if normalized_ioc_type in ["SHA-256", "SHA-1", "MD5"]:
            pattern_type = "file"
            value = f"hashes.'{normalized_ioc_type}'"

        if confidence > 100 or confidence < 0:
            raise MsticpyUserError("confidence must be between 0 and 100")
        data_items = {
            "displayName": name,
            "confidence": confidence,
            "pattern": f"[{pattern_type}:{value} = '{indicator}']",
            "patternType": pattern_type,
            "revoked": "false",
            "source": "MSTICPy",
        }
        data_items.update(_build_additional_indicator_items(**kwargs))
        data = _build_sent_data(data_items, props=True)
        data["kind"] = "indicator"
        response = httpx.post(
            ti_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        if not silent:
            print("Indicator created.")

        return response.json().get("name")

    def bulk_create_indicators(
        self,
        data: pd.DataFrame,
        indicator_column: str = "Observable",
        indicator_type_column: str = "IoCType",
        **kwargs,
    ):
        """
        Bulk create indicators from a DataFrame.

        Parameters
        ----------
        data : pd.DataFrame
            A dataframe containing indicators and indicator types
        indicator_column : str, optional
            The column containing indicator values to create, by default "Observable"
        indicator_type_column : str, optional
            The column containing indicator type values, by default "IoCType"
        confidence_column : str, optional
            The column containing indicator confidence values, by default 0 value used.

        """
        for row in data.iterrows():
            confidence = (
                row[1][kwargs["confidence_column"]]
                if "confidence_column" in kwargs
                else 0
            )
            try:
                self.create_indicator(
                    indicator=row[1][indicator_column],
                    ioc_type=row[1][indicator_type_column],
                    confidence=confidence,
                    silent=True,
                )
            except CloudError:
                print(f"Error creating indicator {row[1][indicator_column]}")
        print(f"{len(data.index)} indicators created.")

    def get_indicator(self, indicator_id: str) -> dict:
        """
        Get a specific indicator by its ID.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to get

        Returns
        -------
        dict
            Indicator details

        Raises
        ------
        CloudError
            If API call fails.

        """
        self.check_connected()  # type: ignore
        ti_url = self.sent_urls["ti"] + f"/indicators/{indicator_id}"  # type: ignore
        params = {"api-version": "2021-10-01"}
        response = httpx.get(
            ti_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            timeout=get_http_timeout(),
        )
        if response.status_code != 200:
            raise CloudError(response=response)
        return response.json()

    def update_indicator(self, indicator_id: str, **kwargs):
        """
        Update an existing indicator within the Microsoft Sentinel workspace.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to update
        name : str, optional
            A common name to give to the indicator default is 'TI Indicator'
        confidence : int, optional
            A score between 0-100 of the confidence in the indicator
        description : str, optional
            An description of the indicator
        labels : list, optional
            A list of string object labels to associate with
            the indicator
        kill_chain_phases : list, optional
            A list of string objects relating to the kill chain phases
            an indicator is assocaited with
        threat_types : list, optional
            A list of threat types associated with the indicator (list of string objects)
        external_references : list, optional
            A list of URLs that provide an external reference for the indicator
        valid_from : datetime, optional
            A datetime from which the indicator is valid from, defaults to now
        valid_to : datetime, optional
            A datetime to which the indicator is valid until

        Raises
        ------
        CloudError
            If API call fails

        """
        self.check_connected()  # type: ignore
        ti_url = self.sent_urls["ti"] + f"/indicators/{indicator_id}"  # type: ignore
        indicator_details = self.get_indicator(indicator_id)
        data_items = _build_additional_indicator_items(**kwargs)
        data_items.pop("validFrom")
        full_data_items = _add_missing_items(data_items, indicator_details)
        data = _build_sent_data(full_data_items, props=True)
        data["etag"] = indicator_details["etag"]
        data["kind"] = "indicator"
        params = {"api-version": "2021-10-01"}
        response = httpx.put(
            ti_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        print("Indicator updated.")

    def add_tag(self, indicator_id: str, tag: str):
        """
        Add a tag to an existing indicator.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to add a tag to.
        tag : str
            The tag to add.

        """
        self.check_connected()  # type: ignore
        indicator_details = self.get_indicator(indicator_id)
        tags = [tag]
        if "threatIntelligenceTags" in indicator_details["properties"]:
            tags += indicator_details["properties"]["threatIntelligenceTags"]
        self.update_indicator(indicator_id=indicator_id, labels=tags)

    def delete_indicator(self, indicator_id: str):
        """
        Delete a specific TI indicator.

        Parameters
        ----------
        indicator_id : str
            The GUID of the indicator to delete

        Raises
        ------
        CloudError
            If API call fails

        """
        self.check_connected()  # type: ignore
        ti_url = self.sent_urls["ti"] + f"/indicators/{indicator_id}"  # type: ignore
        params = {"api-version": "2021-10-01"}
        response = httpx.delete(
            ti_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 204):
            raise CloudError(response=response)
        print("Indicator deleted.")

    def query_indicators(self, **kwargs) -> pd.DataFrame:
        """
        Query for indicators in a Sentinel workspace.

        Parameters
        ----------
        includeDisabled : bool, optional
            Parameter to include/exclude disabled indicators.
        keywords : str, optional
            Keyword for searching threat intelligence indicators
            Use this to search for specific indicator values.
        maxConfidence : int, optional
            Maximum confidence.
        maxValidUntil : str, optional
            End time for ValidUntil filter.
        minConfidence : int, optional
            Minimum confidence.
        minValidUntil : str, optional
            Start time for ValidUntil filter.
        pageSize : int, optional
            Maximum number of results to return in one page.
        patternTypes : list, optional
            A list of IoC types to include.
        sortBy : List, optional
            Columns to sort by and sorting order as:
            [{"itemKey": COLUMN_NAME, "sortOrder": ascending/descending}]
        sources: list, optional
            A list of indicator sources to include
        threatTypes: list, optional
            A list of Threat types to include

        Returns
        -------
        pd.DataFrame
            A set of matching indicators

        Raises
        ------
        CloudError
            If API call fails

        """
        self.check_connected()  # type: ignore
        ti_url = self.sent_urls["ti"] + "/queryIndicators"  # type: ignore
        data_items = dict(kwargs)
        params = {"api-version": "2021-10-01"}
        response = httpx.post(
            ti_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data_items),
            timeout=get_http_timeout(),
        )
        if response.status_code != 200:
            raise CloudError(response=response)
        return _azs_api_result_to_df(response)


def _build_additional_indicator_items(**kwargs) -> dict:
    """Add in additional data items for indicators."""
    data_items = {
        "validFrom": (
            kwargs["valid_from"].isoformat()
            if "valid_from" in kwargs
            else datetime.now().isoformat()
        )
    }
    for item, value in kwargs.items():
        if item in _INDICATOR_ITEMS:
            data_items[_INDICATOR_ITEMS[item]] = value
    if "valid_to" in kwargs:
        data_items["validUntil"] = kwargs["valid_to"].isoformat()
    else:
        data_items["validUntil"] = datetime.now().isoformat()
    if "external_references" in kwargs:
        ext_refs = [
            {"sourceName": "MSTICPy", "url": ref}
            for ref in kwargs["external_references"]
        ]
        data_items["externalReferences"] = ext_refs
    if "kill_chain_phases" in kwargs:
        kill_chain = [
            {
                "killChainName": "Lockheed Martin - Intrusion Kill Chain",
                "phaseName": phase,
            }
            for phase in kwargs["kill_chain_phases"]
        ]
        data_items["killChainPhases"] = kill_chain
    if "labels" in kwargs:
        data_items["labels"] = kwargs["labels"]
        data_items["threatIntelligenceTags"] = kwargs["labels"]
    return data_items


_REQUIRED_ITEMS = ["pattern", "patternType", "source"]


def _add_missing_items(data_items, indicator) -> dict:
    """Add missing required items to requests based on existing values."""
    for req_item in _REQUIRED_ITEMS:
        if req_item not in data_items:
            data_items[req_item] = indicator["properties"][req_item]
    return data_items



#  -------------------------------------------------------------------------
#  Copyright (c) juju4. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for
#  license information.
#  --------------------------------------------------------------------------
"""Sumologic Driver class."""
import re
import time
from datetime import datetime, timedelta
from timeit import default_timer as timer
from typing import Any, Dict, Optional, Tuple, Union

import httpx
import pandas as pd
from sumologic.sumologic import SumoLogic

from ..._version import VERSION
from ...common.exceptions import (
    MsticpyConnectionError,
    MsticpyUserConfigError,
    MsticpyUserError,
)
from ...common.provider_settings import ProviderSettings, get_provider_settings
from ...common.utility import check_kwargs, export
from .driver_base import DriverBase, DriverProps, QuerySource

__version__ = VERSION
__author__ = "juju4"


SUMOLOGIC_CONNECT_ARGS = {
    "connection_str": (
        "(string) The url endpoint (the default is"
        + " 'https://api.us2.sumologic.com/api')."
    ),
    "accessid": (
        "(string) The Sumologic accessid, which is used to "
        + "authenticate on Sumologic instance."
    ),
    "accesskey": "(string) The matching Sumologic accesskey.",
}

_HELP_URI = "https://msticpy.readthedocs.io/en/latest/DataProviders.html"
_SL_NB_URI = (
    "https://github.com/microsoft/msticpy/blob/pr/116-sumologic"
    "-driver-2021-04-21/docs/notebooks/Sumologic-DataConnector.ipynb"
)


@export
class SumologicDriver(DriverBase):
    """Driver to connect and query from Sumologic."""

    _SUMOLOGIC_REQD_ARGS = ["connection_str", "accessid", "accesskey"]
    _CONNECT_DEFAULTS: Dict[str, Any] = {
        "connection_str": "https://api.us2.sumologic.com/api"
    }
    _TIME_FORMAT = '"%Y-%m-%d %H:%M:%S.%6N"'
    _DEF_CHECKINTERVAL = 3
    _DEF_TIMEOUT = 300

    def __init__(self, **kwargs):
        """Instantiate Sumologic Driver."""
        super().__init__(**kwargs)
        self.service = None
        self._loaded = True
        self._connected = False
        self._debug = kwargs.get("debug", False)
        self.set_driver_property(DriverProps.PUBLIC_ATTRS, {"client": self.service})
        self.set_driver_property(
            DriverProps.FORMATTERS, {"datetime": self._format_datetime}
        )
        self.checkinterval = self._DEF_CHECKINTERVAL
        self.timeout = self._DEF_TIMEOUT

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to Sumologic via sumologic-sdk.

        Parameters
        ----------
        connection_str :
            Sumologic API url endpoint. default: https://api.us2.sumologic.com/api

        Other Parameters
        ----------------
        kwargs :
            Connection parameters can be supplied as keyword parameters.

        Notes
        -----
        Default configuration is read from the DataProviders/Sumologic
        section of msticpyconfig.yaml, if available.

        """
        cs_dict = self._get_connect_args(connection_str, **kwargs)

        arg_dict = {
            key: val for key, val in cs_dict.items() if key in SUMOLOGIC_CONNECT_ARGS
        }
        try:
            # https://github.com/SumoLogic/sumologic-python-sdk/blob/master/scripts/search-job.py
            self.service = SumoLogic(
                accessId=arg_dict["accessid"],
                accessKey=arg_dict["accesskey"],
                endpoint=arg_dict["connection_str"],
            )
        except httpx.ConnectError as err:
            raise MsticpyConnectionError(
                f"Authentication error connecting to Sumologic: {err}",
                title="Sumologic connection",
                help_uri=_HELP_URI,
                nb_uri=_SL_NB_URI,
            ) from err
        except httpx.HTTPError as err:
            raise MsticpyConnectionError(
                f"Communication error connecting to Sumologic: {err}",
                title="Sumologic connection",
                help_uri=_HELP_URI,
                nb_uri=_SL_NB_URI,
            ) from err
        except Exception as err:
            raise MsticpyConnectionError(
                f"Error connecting to Sumologic: {err}",
                title="Sumologic connection",
                help_uri=_HELP_URI,
                nb_uri=_SL_NB_URI,
            ) from err
        self._connected = True
        print(f"connected with accessid {arg_dict['accessid']}")

    def _get_connect_args(
        self, connection_str: Optional[str], **kwargs
    ) -> Dict[str, Any]:
        """Check and consolidate connection parameters."""
        cs_dict: Dict[str, Any] = self._CONNECT_DEFAULTS
        # Fetch any config settings
        settings, cs_is_instance_name = self._get_sumologic_settings(connection_str)
        cs_dict.update(settings)
        # If a connection string - parse this and add to config
        if connection_str and not cs_is_instance_name:
            cs_dict["connection_str"] = connection_str
        if kwargs:
            # if connection args supplied as kwargs
            cs_dict.update(kwargs)
            check_kwargs(cs_dict, list(SUMOLOGIC_CONNECT_ARGS.keys()))

        missing_args = set(self._SUMOLOGIC_REQD_ARGS) - cs_dict.keys()
        if missing_args:
            raise MsticpyUserConfigError(
                "One or more connection parameters missing for Sumologic connector",
                ", ".join(missing_args),
                f"Required parameters are {', '.join(self._SUMOLOGIC_REQD_ARGS)}",
                "All parameters:",
                *[f"{arg}: {desc}" for arg, desc in SUMOLOGIC_CONNECT_ARGS.items()],
                title="no Sumologic connection parameters",
                help_uri=_HELP_URI,
                notebook_uri=_SL_NB_URI,
            )
        return cs_dict

    # pylint: disable=broad-except
    def _query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute Sumologic query and retrieve results.

        Parameters
        ----------
        query : str
            Sumologic query to execute
        query_source : QuerySource
            Not used.

        Other Parameters
        ----------------
        days : int
            Search the past X days.
        start : datetime
            A datetime() object representing the start of the search
            window. If used without end_time, the end of the search
            window is the current time.
        start_time : datetime
            alias for `start`
        end : datetime
            A datetime() object representing the end of the search window.
            If used without start_time, the search start will be the earliest
            time in the index.
        end_time:
            alias for `end`
        timezone : str
            timezone used for time range search
        byreceipttime : datetime
            if time reference should used _receiptTime (time when Sumologic
            got message) instead of _messageTime (time present in log message).
        limit : int
            An integer describing the max number of search results to return.
        forcemessagesresults : bool
            Force results to be raw messages even if aggregated query.
        verbosity : int
            Provide more verbose state. from 0 least verbose to 4 most one.
        checkinterval : int
            interval in seconds to check if results are gathered
        timeout : int
            timeout in seconds when gathering results

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err("SumoLogic")

        verbosity = kwargs.pop("verbosity", 0)
        timezone = kwargs.pop("timezone", "UTC")
        by_receipt_time = kwargs.pop("byreceipttime", False)
        self.checkinterval = kwargs.pop("checkinterval", self._DEF_CHECKINTERVAL)
        self.timeout = kwargs.pop("timeout", self._DEF_TIMEOUT)

        start_time, end_time = self._get_time_params(**kwargs)

        # default to unlimited query unless count is specified
        # https://help.sumologic.com/05Search/Search-Query-Language/Search-Operators/limit
        if "limit" in kwargs and kwargs["limit"] <= 10000:
            limit = kwargs["limit"]
            query = f"{query} | limit {limit}"
        else:
            limit = None

        if verbosity >= 1:
            print(f"INFO: from {start_time} to {end_time}, TZ {timezone}")

        if verbosity >= 2:
            print(f"DEBUG: query {query}")
            print(f"DEBUG: byreceipttime {by_receipt_time}")
            timer_start = timer()
        # submit the search job
        try:
            searchjob = self.service.search_job(
                query, start_time, end_time, timezone, by_receipt_time
            )
        except Exception as err:
            self._raise_qry_except(err, "submit search_job", "to submit search job")

        qry_count = re.search(r"\|\s*count", query, re.IGNORECASE)
        if verbosity >= 2:
            print(f"DEBUG: search job {searchjob}")
            print(f"DEBUG: search performance: {timer() - timer_start}")
            print(f"DEBUG: messages or records? {qry_count}")

        # poll for job completion
        status = self._poll_job_status(searchjob, verbosity)
        print(status["state"])

        # return the results
        return self._get_job_results(
            searchjob=searchjob,
            status=status,
            qry_count=qry_count,
            force_mssg_rstls=kwargs.pop("forcemessagesresults", False),
            limit=limit,
            verbosity=verbosity,
        )

    def _poll_job_status(self, searchjob, verbosity):
        status = self.service.search_job_status(searchjob)
        if verbosity >= 2:
            print(f"DEBUG: status {status}")
        time_counter = 0
        while status["state"] != "DONE GATHERING RESULTS":
            if status["state"] == "CANCELLED":
                break
            status = self.service.search_job_status(searchjob)
            if verbosity >= 4:
                print(
                    f"DEBUG: pending results, state {status['state']}.",
                    f"slept {time_counter}s. sleeping extra {self.checkinterval}s",
                    "until {self.timeout}s",
                )
            if time_counter < self.timeout:
                time.sleep(self.checkinterval)
                time_counter += self.checkinterval
            else:
                print(
                    f"WARN: wait more than timeout {self.timeout}. stopping. "
                    + "Use timeout argument to wait longer."
                )
                break
        return status

    # pylint: disable=inconsistent-return-statements
    def _get_job_results_messages(self, searchjob, status, limit):
        # Non-aggregated results, Messages only
        count = status["messageCount"]
        limit2 = None
        if limit is not None:
            limit2 = (
                count if count < limit and count != 0 else limit
            )  # compensate bad limit check
        try:
            result = self.service.search_job_messages(searchjob, limit=limit2)
            return result["messages"]
        except Exception as err:
            self._raise_qry_except(err, "search_job_messages", "to get job messages")

    def _get_job_results_records(  # noqa: MC0001
        self, searchjob, status, limit, verbosity
    ):
        # Aggregated results, limit
        count = status["recordCount"]
        limit2 = None
        if limit is not None:
            limit2 = (
                count if count < limit and count != 0 else limit
            )  # compensate bad limit check

            if count < limit:
                if verbosity >= 2:
                    print(f"DEBUG: No Paging, total count {count}, limit {limit}")
                try:
                    result = self.service.search_job_records(searchjob, limit=limit2)
                    return result["records"]
                except Exception as err:
                    self._raise_qry_except(
                        err, "search_job_records", "to get search records"
                    )
        else:
            # paging results
            # https://help.sumologic.com/APIs/Search-Job-API/About-the-Search-Job-API#query-parameters-2
            if verbosity >= 2:
                print(f"DEBUG: Paging, total count {count}, limit {limit}")
            try:
                job_limit = 10000
                iterations = int(count / job_limit) + (count % job_limit > 0)
                total_results = []
                for i in range(0, iterations):
                    if i == iterations:
                        job_limit2 = count - (iterations - 1) * job_limit
                    else:
                        job_limit2 = job_limit
                    if verbosity >= 2:
                        print(
                            f"DEBUG: Paging {i * job_limit} / {count}, limit {job_limit2}"
                        )
                    result = self.service.search_job_records(
                        searchjob, offset=(i * job_limit), limit=job_limit2
                    )
                    total_results.extend(result["records"])
                return total_results
            except Exception as err:
                self._raise_qry_except(
                    err,
                    "search_job_records",
                    f"to get search records (paging i {i*job_limit} / {count})",
                )

    # pylint: disable=inconsistent-return-statements
    # I don't think there are any - everything returns a list
    def _get_job_results(
        self, searchjob, status, qry_count, force_mssg_rstls, limit, verbosity
    ):
        if status["state"] != "DONE GATHERING RESULTS":
            return []
        if not qry_count or force_mssg_rstls:
            # Non-aggregated results, Messages only
            return self._get_job_results_messages(searchjob, status, limit)

        # Aggregated results, limit
        return self._get_job_results_records(searchjob, status, limit, verbosity)

    # pylint: enable=inconsistent-return-statements

    @staticmethod
    def _raise_qry_except(err: Exception, mssg: str, action: Optional[str] = None):
        if isinstance(err, httpx.HTTPError):
            raise MsticpyConnectionError(
                f"Communication error connecting to Sumologic: {err}",
                title=f"Sumologic {mssg}",
                help_uri=_HELP_URI,
                notebook_uri=_SL_NB_URI,
            ) from err
        action = action or mssg
        raise MsticpyConnectionError(
            f"Failed {action}: {err}",
            title=f"Sumologic - {mssg}",
            help_uri=_HELP_URI,
            notebook_uri=_SL_NB_URI,
        ) from err

    def _get_time_params(self, **kwargs):
        if "days" in kwargs:
            end = datetime.now()
            start = end - timedelta(days=kwargs["days"])
            return self._format_datetime(start), self._format_datetime(end)

        start = kwargs.pop("start", kwargs.pop("start_time", None))
        end = kwargs.pop("end", kwargs.pop("end_time", None))
        if start and not end:
            end = datetime.now()
        elif not start:
            raise MsticpyUserError(
                "Error! requires either 'days' or 'start' parameters",
                title="Missing parameter.",
                help_uri=_HELP_URI,
                notebook_uri=_SL_NB_URI,
            )
        return self._format_datetime(start), self._format_datetime(end)

    # pylint: disable=too-many-branches
    def query(  # noqa: MC0001
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute Sumologic query and retrieve results.

        Parameters
        ----------
        query : str
            Sumologic query to execute
        query_source : QuerySource
            Not used.

        Other Parameters
        ----------------
        days: int
            Search the past X days.
        start : datetime
            A datetime() object representing the start of the search
            window. If used without end_time, the end of the search
            window is the current time.
        start_time : datetime
            alias for `start`
        end : datetime
            A datetime() object representing the end of the search window.
            If used without start_time, the search start will be the earliest
            time in the index.
        end_time : datetime
            alias for `end`
        timeZone : str
            timezone used for time range search
        byReceiptTime : datetime
            if time reference should used _receiptTime (time when Sumologic
            got message) instead of _messageTime (time present in log message).
        limit : int
            An integer describing the max number of search results to return.
        forceMessagesResults : bool
            Force results to be raw messages even if aggregated query.
        verbosity : int
            Provide more verbose state. from 0 least verbose to 4 most one.
        normalize : bool
            If set to True, fields containing structures (i.e. subfields)
            will be flattened such that each field has it's own column in
            the dataframe. If False, there will be a single column for the
            structure, with a JSON string encoding all the contents.
        exporting : bool
            Export result to file.
        export_path : str
            file path for exporte results.
        time_columns: array[string]
            returning columns which format should be dataframe timestamp
        numeric_columns: array[string]
            returning columns which format should be dataframe numeric

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        limit = kwargs.get("limit", None)
        verbosity = kwargs.get("verbosity", 0)
        normalize = kwargs.pop("normalize", True)
        exporting = kwargs.pop("exporting", False)
        export_path = kwargs.pop("export_path", "")
        time_columns = kwargs.pop("time_columns", [])
        numeric_columns = kwargs.pop("numeric_columns", [])

        results = self._query(query, **kwargs)
        if verbosity >= 3:
            print("DEBUG: {results}")
        if normalize:
            dataframe_res = pd.json_normalize(results)  # type: ignore
        else:
            dataframe_res = pd.DataFrame(results)

        if limit is not None and dataframe_res.shape[0] > limit:
            dataframe_res = dataframe_res.head(limit)

        for col in dataframe_res.columns:
            try:
                if (
                    col
                    in ["map._count", "map._collectorid", "map._messageid", "map._size"]
                    + numeric_columns
                ):
                    dataframe_res[col] = pd.to_numeric(dataframe_res[col])
                # ensure timestamp format
                # https://help.sumologic.com/05Search/Get-Started-with-Search/Search-Basics/Built-in-Metadata
                # https://help.sumologic.com/05Search/Search-Query-Language/Search-Operators/timeslice
                if col in ("map._receipttime", "map._messagetime", "map._timeslice"):
                    dataframe_res[col] = pd.to_datetime(dataframe_res[col], unit="ms")
                if col in time_columns:
                    dataframe_res[col] = pd.to_datetime(dataframe_res[col])

            except Exception as err:
                self._raise_qry_except(
                    err,
                    "query",
                    f"query column type conversion: {col} -> {dataframe_res[col]}",
                )

        if exporting:
            if export_path.endswith(".xlsx"):
                if verbosity >= 2:
                    print(f"DEBUG: Exporting results to excel file {export_path}")
                dataframe_res.to_excel(export_path, index=False)
            elif export_path.endswith(".csv"):
                if verbosity >= 2:
                    print("DEBUG: Exporting results to csv file {export_path}")
                dataframe_res.to_csv(export_path, index=False)

        return dataframe_res.copy()

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            Query to execute against Sumologic instance.

        Returns
        -------
        Union[pd.DataFrame,Any]
            A DataFrame (if successful) or
            the underlying provider result if an error occurs.

        """
        raise NotImplementedError(f"Not supported for {self.__class__.__name__}")

    # Parameter Formatting methods
    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return date_time.strftime("%Y-%m-%dT%H:%M:%S")

    # Read values from configuration
    @staticmethod
    def _get_sumologic_settings(
        instance_name: str = None,
    ) -> Tuple[Dict[str, Any], bool]:
        """Get config from msticpyconfig."""
        data_provs = get_provider_settings(config_section="DataProviders")
        sl_settings = {
            name: settings
            for name, settings in data_provs.items()
            if name.startswith("Sumologic")
        }
        sumologic_settings: Optional[ProviderSettings]
        # Check if the connection string is an instance name
        sumologic_settings = sl_settings.get(f"Sumologic-{instance_name}")
        if sumologic_settings:
            is_instance_name = True
        else:
            # otherwise get the default Sumologic entry
            sumologic_settings = sl_settings.get("Sumologic")
            is_instance_name = False
        return getattr(sumologic_settings, "args", {}), is_instance_name



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query Editor."""

from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Dict, Optional, Union, cast

import ipywidgets as widgets
import yaml
from IPython.display import display

from .._version import VERSION
from ..data.core.query_defns import DataEnvironment, DataFamily
from ..data.core.query_template import (
    Query,
    QueryArgs,
    QueryCollection,
    QueryDefaults,
    QueryMetadata,
    QueryParameter,
)

__version__ = VERSION
__author__ = "Ian Hellen"

# pylint: disable=too-many-instance-attributes, too-many-lines


class IPyDisplayMixin:
    """IPython display mixin class."""

    def display(self):
        """Display the interactive widgets."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


_PARAM_OPTIONS = ("str", "datetime", "list", "int")
_DESC_WIDTH = "120px"


def txt_fmt(width: str = "70%") -> Dict:
    """Return a dictionary with layout and style for a text widget.

    Parameters
    ----------
    width : str, optional
        The width of the widget, by default "70%"
    height : str, optional
        The height of the widget, by default "20px"

    Returns
    -------
    Dict
        A dictionary with layout and style for a text widget.

    """
    return {
        "layout": widgets.Layout(width=width),
        "style": {"description_width": _DESC_WIDTH},
    }


def txtarea_fmt(width: str = "70%", height: str = "150px") -> Dict:
    """Return a dictionary with layout and style for a textarea widget.

    Parameters
    ----------
    width : str, optional
        The width of the widget, by default "70%"
    height : str, optional
        The height of the widget, by default "150px"

    Returns
    -------
    Dict
        A dictionary with layout and style for a textarea widget.

    """
    return {
        "layout": widgets.Layout(width=width, height=height),
        "style": {"description_width": _DESC_WIDTH},
    }


def sel_fmt(width: str = "40%", height: str = "150px") -> Dict:
    """Return a dictionary with layout and style for a select widget.

    Parameters
    ----------
    width : str, optional
        The width of the widget, by default "40%"
    height : str, optional
        The height of the widget, by default "150px"

    Returns
    -------
    Dict
        A dictionary with layout and style for a select widget.

    """
    return {
        "layout": widgets.Layout(width=width, height=height),
        "style": {"description_width": _DESC_WIDTH},
    }


def box_layout():
    """Return a dictionary with layout and style for a box widget.

    Returns
    -------
    Dict
        A dictionary with layout and style for a select widget.

    """
    return {
        "layout": widgets.Layout(
            border="1px solid black",
            margin="5px",
            padding="5px",
            style={"padding": "5px", "background-color": "red"},
        )
    }


@dataclass
class CustomChange:
    """Dummy ipywidgets change event."""

    new: Any


class QueryParameterEditWidget(IPyDisplayMixin):
    """Class to manage editing of query parameters.

    This class provides a graphical user interface for editing query parameters.
    It allows users to add, modify, and delete parameters, as well as specify
    their names, descriptions, types, default values, and whether they are mandatory.

    Attributes
    ----------
    _changed_data : bool
        A flag indicating whether the data has been changed.
    param_container : Union[Query, QueryDefaults]
        The container for the query parameters.
    parameter_dropdown : ipywidgets.Select
        A widget for selecting a parameter to edit.
    parameter_name_widget : ipywidgets.Text
        A widget for editing the name of a parameter.
    description_widget : ipywidgets.Text
        A widget for editing the description of a parameter.
    type_widget : ipywidgets.Dropdown
        A widget for selecting the type of a parameter.
    default_reqd_widget : ipywidgets.Checkbox
        A widget for indicating whether a default value is required for a parameter.

    """

    def __init__(self, container: Union[Query, QueryDefaults]):
        """Initialize the class."""
        self._changed_data = False
        self.param_container = container
        self.parameter_dropdown = widgets.Select(
            description="Parameters",
            size=5,
            options=list(
                self.param_container.parameters.keys()
                if self.param_container.parameters
                else []
            ),
            **sel_fmt(height="100px"),
        )
        # Create widgets for the Parameter fields
        self.parameter_name_widget = widgets.Text(description="Name", **txt_fmt())
        self.description_widget = widgets.Text(description="Description", **txt_fmt())
        self.type_widget = widgets.Dropdown(
            description="Type", options=_PARAM_OPTIONS, **sel_fmt(height="30px")
        )
        self.default_reqd_widget = widgets.Checkbox(description="Use a default value")
        self.default_widget = widgets.Text(description="Default Value", **txt_fmt())

        # Create buttons
        self.add_parameter_button = widgets.Button(description="New Parameter")
        self.save_parameter_button = widgets.Button(description="Save Parameter")
        self.delete_parameter_button = widgets.Button(description="Delete Parameter")

        # Attach the functions to buttons
        self.add_parameter_button.on_click(self.add_parameter)
        self.save_parameter_button.on_click(self.save_parameter)
        self.delete_parameter_button.on_click(self.delete_parameter)
        self.parameter_dropdown.observe(self.populate_widgets, names="value")

        # Create a widget for adding, editing, and deleting Parameters
        self.layout = widgets.VBox(
            [
                widgets.HBox(
                    [
                        self.parameter_dropdown,
                        widgets.VBox(
                            [
                                self.add_parameter_button,
                                self.delete_parameter_button,
                            ]
                        ),
                    ]
                ),
                widgets.VBox(
                    children=[
                        self.parameter_name_widget,
                        self.description_widget,
                        self.type_widget,
                        widgets.HBox([self.default_reqd_widget, self.default_widget]),
                        self.save_parameter_button,
                    ],
                    **box_layout(),
                ),
            ]
        )
        if self.param_container and self.param_container.parameters:
            init_change = CustomChange(new=next(iter(self.param_container.parameters)))
            self.populate_widgets(init_change)

    @property
    def changed_data(self):
        """Return True if data has changed."""
        return self._changed_data

    def reset_changed_data(self):
        """Reset changed data flag."""
        self._changed_data = False

    def set_param_container(self, container: Union[Query, QueryDefaults]):
        """Set the parameter container."""
        self.param_container = container
        if self.param_container and self.param_container.parameters:
            self.parameter_dropdown.options = list(
                self.param_container.parameters.keys()
            )
            init_change = CustomChange(new=next(iter(self.param_container.parameters)))
            self.populate_widgets(init_change)
        else:
            self.parameter_dropdown.options = []
            self._blank_parameter()

    # Define a function to add a new Parameter to the selected Query
    def add_parameter(self, button):
        """Add a new parameter."""
        del button
        # Clear the input widgets
        self._blank_parameter()
        self.parameter_name_widget.value = "new_parameter"

    def _blank_parameter(self):
        """Clear the parameter widgets."""
        self.parameter_name_widget.value = ""
        self.description_widget.value = ""
        self.type_widget.value = _PARAM_OPTIONS[0]
        self.default_widget.value = ""
        self.default_reqd_widget.value = False

    # Define a function to populate the Parameter widgets with the values of the selected Parameter
    def populate_widgets(self, change):
        """Populate parameter value in widgets."""
        parameter = self.param_container.parameters[change.new]
        self.parameter_name_widget.value = change.new
        self.description_widget.value = parameter.description
        self.type_widget.value = parameter.datatype
        self.default_reqd_widget.value = parameter.default is not None
        self.default_widget.value = parameter.default or ""

    # Define a function to edit the selected Parameter with the values from the widgets
    def save_parameter(self, button):
        """Save currently edited parameter."""
        del button
        if not self.parameter_name_widget.value:
            return
        param_name = self.parameter_name_widget.value
        parameter = QueryParameter(
            description=self.description_widget.value,
            datatype=self.type_widget.value or _PARAM_OPTIONS[0],
        )
        parameter.default = (
            self.default_widget.value if self.default_reqd_widget.value else None
        )
        self.param_container.parameters[param_name] = parameter
        self.parameter_dropdown.options = list(self.param_container.parameters.keys())
        self.parameter_dropdown.value = param_name
        self._changed_data = True

    # Define a function to delete the selected Parameter from the selected Query
    def delete_parameter(self, button):
        """Delete parameter item."""
        del button
        del self.param_container.parameters[self.parameter_dropdown.value]
        # Clear the input widgets
        self._blank_parameter()
        self._changed_data = True


def replace_in_query(
    param_name: str, param: QueryParameter, src_name: str, query: str
) -> str:
    """
    Replace a parameter in a query string with a formatted string.

    Parameters
    ----------
    param_name : str
        The name of the parameter to replace.
    param : Parameter
        The parameter object containing the datatype of the parameter.
    src_name : str
        The name of the source to replace in the query string.
    query : str
        The query string to replace the source in.

    Returns
    -------
    str
        The query string with the source replaced by the formatted parameter string.

    """
    if param.datatype == "datetime":  # type: ignore
        repl_str = f"datetime({{{param_name}}})"
    elif param.datatype == "str":
        repl_str = f'"{{{param_name}}}"'
    else:
        repl_str = f"{{{param_name}}}"
    return query.replace(src_name, repl_str)


class QueryEditWidget(IPyDisplayMixin):
    """A class for editing queries."""

    def __init__(self, query_collection: QueryCollection):
        """
        Initialize a QueryEditWidget object.

        Parameters
        ----------
        query_collection : QueryCollection
            A collection of queries.

        """
        self._changed_data = False
        self.query_collection = query_collection
        self.query_select_widget = widgets.Select(
            description="Queries",
            options=list(self.query_collection.sources.keys()),
            **sel_fmt(),
        )
        self.name_widget = widgets.Text(
            description="Name", placeholder="Python-compatible query name", **txt_fmt()
        )
        self.description_widget = widgets.Text(
            description="Description",
            placeholder="Description of query",
            **txt_fmt("90%"),
        )
        self.metadata_widget = widgets.Textarea(
            description="Query metadata", **txtarea_fmt("90%", height="70px")
        )
        self.parameters: Optional[QueryParameterEditWidget] = None
        self.query_text_widget = widgets.Textarea(
            description="Query",
            placeholder="query text with {parameter}a",
            **txtarea_fmt("90%"),
        )
        self.query_opts_widget = widgets.Accordion(
            children=[
                widgets.Label(value="No parameters"),
                self.metadata_widget,
            ],
            titles=["Query parameters", "Query metadata"],
            layout=widgets.Layout(margin="5px"),
        )
        self.query_opts_widget.set_title(0, "Query parameters")
        self.query_opts_widget.set_title(1, "Query metadata")
        self.query_opts_widget.selected_index = None
        self.add_query_button = widgets.Button(description="New Query")
        self.save_query_button = widgets.Button(description="Save Query")
        self.delete_query_button = widgets.Button(description="Delete Query")
        self.queries_widget = widgets.VBox(
            children=[
                widgets.Label(value="Query"),
                self.name_widget,
                self.description_widget,
                self.query_text_widget,
                self.query_opts_widget,
                self.save_query_button,
            ],
            **box_layout(),
        )
        self.layout = widgets.VBox(
            [
                widgets.Label(value="Queries"),
                widgets.HBox(
                    children=[
                        self.query_select_widget,
                        widgets.VBox(
                            children=[
                                self.add_query_button,
                                self.delete_query_button,
                            ],
                        ),
                    ],
                    **box_layout(),
                ),
                self.queries_widget,
            ]
        )
        self.add_query_button.on_click(self.add_query)
        self.query_select_widget.observe(self.populate_widgets, names="value")
        self.save_query_button.on_click(self.save_query)
        self.delete_query_button.on_click(self.delete_query)
        self.populate_query_dropdown(None)
        if self.query_collection.sources:
            init_change = CustomChange(new=next(iter(self.query_collection.sources)))
            self.populate_widgets(init_change)

    @property
    def changed_data(self):
        """Return True if data has changed."""
        return self._changed_data or (
            self.parameters.changed_data if self.parameters else False
        )

    def reset_changed_data(self):
        """Reset changed data flag."""
        self._changed_data = False
        if self.parameters:
            self.parameters.reset_changed_data()

    def set_query_collection(self, query_collection: QueryCollection):
        """Set the query collection."""
        self.query_collection = query_collection
        self.populate_query_dropdown(None)
        if self.query_collection.sources:
            init_change = CustomChange(new=next(iter(self.query_collection.sources)))
            self.populate_widgets(init_change)

    def populate_query_dropdown(self, change: Any) -> None:
        """
        Populate the query dropdown widget.

        Parameters
        ----------
        change : Any
            The change event.

        """
        del change
        self.query_select_widget.options = list(self.query_collection.sources.keys())

    def _populate_qry_metadata(self, query: Query) -> str:
        """
        Populate the metadata widget with the metadata of a query.

        Parameters
        ----------
        query : Query
            The query object.

        """
        return yaml.safe_dump(query.metadata) if query.metadata else ""

    def _save_qry_metadata(self, metadata: str) -> Dict[str, Any]:
        """
        Save the metadata of a query.

        Parameters
        ----------
        metadata : str
            The metadata string.

        """
        return yaml.safe_load(metadata) if metadata else {}

    def populate_widgets(self, change):
        """
        Populate the query widgets with the data of a query.

        Parameters
        ----------
        change : Any
            The change event.

        """
        query = self.query_collection.sources[change.new]
        self.name_widget.value = change.new
        self.description_widget.value = query.description
        self.metadata_widget.value = self._populate_qry_metadata(query)
        self.query_text_widget.value = self._fmt_query(query.args.query)
        self.parameters = QueryParameterEditWidget(query)
        self.query_opts_widget.children = [
            self.parameters.layout,
            self.metadata_widget,
        ]

    @staticmethod
    def _fmt_query(query: str) -> str:
        """
        Format a query string.

        Parameters
        ----------
        query : str
            The query string.

        Returns
        -------
        str
            The formatted query string.

        """
        return "\n|".join(query.strip().split("|"))

    def add_query(self, button):
        """
        Add a new query.

        Parameters
        ----------
        button : Button
            The button object.

        """
        del button
        self.name_widget.value = ""
        self.description_widget.value = ""
        self.metadata_widget.value = ""
        self.query_text_widget.value = ""

    def save_query(self, button):
        """
        Save the data of a query.

        Parameters
        ----------
        button : Button
            The button object.

        """
        del button
        if self.name_widget.value not in self.query_collection.sources:
            self.query_collection.sources[self.name_widget.value] = Query(
                description=self.description_widget.value,
                metadata=self._save_qry_metadata(self.metadata_widget.value),
                args=QueryArgs(query=self.query_text_widget.value),
                parameters={},
            )
            self.populate_query_dropdown(None)
        else:
            query = self.query_collection.sources[self.name_widget.value]
            query = self.query_collection.sources[self.name_widget.value]
            query.description = self.description_widget.value
            query.metadata = self._save_qry_metadata(self.metadata_widget.value)
            query.args = QueryArgs(query=self.query_text_widget.value)
        self._changed_data = True

    def delete_query(self, button):
        """
        Delete a query.

        Parameters
        ----------
        button : Button
            The button object.

        """
        del button
        del self.query_collection.sources[self.query_select_widget.value]
        self.name_widget.value = ""
        self.description_widget.value = ""
        self.metadata_widget.value = ""
        self.query_text_widget.value = ""
        self.populate_query_dropdown(None)
        self._changed_data = True


class MetadataEditWidget(IPyDisplayMixin):
    """A class for editing Metadata properties."""

    def __init__(self, metadata: Optional[QueryMetadata] = None):
        """
        Initialize a MetadataEditWidget object.

        Parameters
        ----------
        metadata : Metadata
            A Metadata object.

        """
        self._changed_data = False
        self.metadata = metadata or self._new_metadata()
        self.version_widget = widgets.IntText(description="Version", **txt_fmt())
        self.description_widget = widgets.Text(description="Description", **txt_fmt())
        self.data_env_widget = widgets.SelectMultiple(
            options=list(DataEnvironment.__members__),
            description="Data Environments",
            **sel_fmt(),
        )
        self.data_families_widget = widgets.Textarea(
            description="Data Families",
            **sel_fmt(),
        )
        self.database_widget = widgets.Text(
            description="Kusto Database", placeholder="Database name", **txt_fmt()
        )
        self.cluster_widget = widgets.Text(
            description="Kusto Cluster",
            placeholder="URI, hostname or alias",
            **txt_fmt(),
        )
        self.clusters_widget = widgets.Textarea(
            description="Kusto Clusters",
            placeholder="URI, hostname or alias (one per line)",
            **sel_fmt(width="70%", height="75px"),
        )
        self.cluster_groups_widget = widgets.Textarea(
            description="Kusto Clusters",
            placeholder="MSTICPy cluster group (one per line)",
            **sel_fmt(width="70%", height="75px"),
        )
        self.tags_widget = widgets.Text(
            description="Tags", placeholder="(optional)", **txt_fmt()
        )
        self.data_source_widget = widgets.Text(
            description="Data Source", placeholder="(optional)", **txt_fmt()
        )
        self.save_metadata_widget = widgets.Button(description="Save metadata")
        self.save_metadata_widget.on_click(self.save_metadata)

        self.layout = widgets.VBox(
            [
                self.version_widget,
                self.description_widget,
                self.data_env_widget,
                self.data_families_widget,
                self.database_widget,
                widgets.Label(
                    "Specify the cluster, clusters or cluster groups for the query.",
                ),
                self.cluster_widget,
                self.clusters_widget,
                self.cluster_groups_widget,
                widgets.Label(
                    "Other items",
                ),
                self.tags_widget,
                self.data_source_widget,
                self.save_metadata_widget,
            ]
        )
        self.populate_widgets()

    @property
    def changed_data(self):
        """Return True if data has changed."""
        return self._changed_data

    def reset_changed_data(self):
        """Reset changed data flag."""
        self._changed_data = False

    def set_metadata(self, metadata: QueryMetadata):
        """Set the metadata object."""
        self.metadata = metadata
        self.populate_widgets()

    def populate_widgets(self):
        """Populate the widgets with the metadata of the Metadata object."""
        self.version_widget.value = self.metadata.version or ""
        self.description_widget.value = self.metadata.description or ""
        self.data_env_widget.value = (
            tuple(self.metadata.data_environments)
            if self.metadata.data_environments
            else ()
        )
        self.data_families_widget.value = (
            ", ".join(self.metadata.data_families)
            if self.metadata.data_families
            else ""
        )
        self.database_widget.value = self.metadata.database or ""
        self.cluster_widget.value = self.metadata.cluster or ""
        self.clusters_widget.value = "\n".join(self.metadata.clusters or [])
        self.cluster_groups_widget.value = "\n".join(self.metadata.cluster_groups or [])
        self.tags_widget.value = (
            ", ".join(self.metadata.tags) if self.metadata.tags else ""
        )
        self.data_source_widget.value = self.metadata.data_source or ""

    def save_metadata(self, button):
        """Save the values to the Metadata object."""
        del button
        self.metadata.version = self.version_widget.value
        self.metadata.description = self.description_widget.value
        self.metadata.data_environments = list(self.data_env_widget.value)
        self.metadata.data_families = [
            fam.strip()
            for fam in self.data_families_widget.value.split(",")
            if fam.strip()
        ]
        self.metadata.database = self.database_widget.value
        self.metadata.cluster = self.cluster_widget.value
        self.metadata.clusters = [
            cluster.strip() for cluster in self.clusters_widget.value.split("\n")
        ]
        self.metadata.cluster_groups = [
            cluster_grp.strip()
            for cluster_grp in self.cluster_groups_widget.value.split("\n")
        ]
        self.metadata.tags = [
            tag.strip() for tag in self.tags_widget.value.split(",") if tag.strip()
        ]
        self.metadata.data_source = self.data_source_widget.value
        self._changed_data = True

    def display(self):
        """Display the GUI widget."""
        display(self.layout)

    def _new_metadata(self):
        return QueryMetadata(
            version=1,
            description="Query collection description",
            data_environments=[DataEnvironment.Unknown.name],
            data_families=[DataFamily.Unknown.name],
        )


_DEF_FILENAME = "new_query_file.yaml"


class QueryEditor(IPyDisplayMixin):
    """Query template editor."""

    def __init__(
        self,
        query_file: Union[QueryCollection, Path, str, None] = None,
    ):
        """
        Initialize the QueryEditor.

        Parameters
        ----------
        query_file : Union[QueryCollection, Path, str, None], optional
            The query collection to edit, can be a string or pathlib Path
            object for a file to load, or a QueryCollection object.
            By default None

        """
        self.filename_widget = widgets.Text(
            description="Current file", layout=widgets.Layout(width="70%")
        )
        if isinstance(query_file, (Path, str)):
            self.filename_widget.value = str(query_file)
            self._open_initial_file()
        else:
            self.query_collection = query_file or self._new_collection()
            self.filename_widget.value = (
                self.query_collection.file_name or _DEF_FILENAME
            )
        self.query_editor = QueryEditWidget(self.query_collection)
        self.metadata_editor = MetadataEditWidget(self.query_collection.metadata)
        if not self.query_collection.defaults:
            self.query_collection.defaults = QueryDefaults(
                metadata={}, parameters={}  # type: ignore[call-arg]
            )
        self.default_param_editor = QueryParameterEditWidget(
            self.query_collection.defaults
        )

        self.ignore_changes = widgets.Checkbox(
            description="Ignore changes", value=False
        )
        self.open_button = widgets.Button(description="Open File")
        self.save_button = widgets.Button(description="Save File")
        self.new_button = widgets.Button(description="New file")
        self.collection_accordion = widgets.Accordion(
            children=[self.metadata_editor.layout, self.default_param_editor.layout],
            titles=["File metadata", "Default parameters"],
        )
        self.collection_accordion.set_title(0, "File metadata")
        self.collection_accordion.set_title(1, "Default parameters")
        self.collection_accordion.selected_index = None
        self.layout = widgets.VBox(
            [
                widgets.HTML("<h1 style='text-indent: 15px'>MSTICPy Query Editor</h1>"),
                self.collection_accordion,
                self.query_editor.layout,
                self._create_file_widget(),
            ]
        )
        self.open_button.on_click(self._open_file)
        self.save_button.on_click(self._save_file)
        self.new_button.on_click(self._new_file)

    @property
    def current_file(self) -> str:
        """Return the current file name."""
        return cast(str, self.filename_widget.value) or cast(str, _DEF_FILENAME)

    def _create_file_widget(self):
        """Create the file widget."""
        return widgets.HBox(
            children=[
                self.filename_widget,
                widgets.VBox(
                    [
                        self.new_button,
                        self.open_button,
                        self.save_button,
                        self.ignore_changes,
                    ],
                    layout=widgets.Layout(width="30%"),
                ),
            ]
        )

    def _new_collection(self):
        """Create a new query collection."""
        return QueryCollection(
            file_name=_DEF_FILENAME,
            metadata=QueryMetadata(
                version=1, description="new", data_families=[], data_environments=[]
            ),
            defaults=QueryDefaults(),
        )

    def _update_query_collection(self, query_collection: QueryCollection):
        """Update the sub-editors with the current query collection."""
        self.query_collection = query_collection
        self.query_editor.set_query_collection(self.query_collection)
        if not self.query_collection.defaults:
            self.query_collection.defaults = QueryDefaults(
                metadata={}, parameters={}  # type: ignore[call-arg]
            )
        self.default_param_editor.set_param_container(self.query_collection.defaults)
        self.metadata_editor.set_metadata(self.query_collection.metadata)

    def _save_file(self, button):
        """Save the current query collection."""
        del button
        save_queries_to_yaml(self.query_collection, self.current_file)

    def _open_file(self, button):
        """Open a new query collection."""
        del button
        if self._unsaved_changes() and not self.ignore_changes.value:
            print(
                "Please save or check 'Ignore changes' before opening a different file."
            )
            return
        self._reset_change_state()
        self.query_collection = load_queries_from_yaml(self.current_file)
        self._update_query_collection(self.query_collection)

    def _open_initial_file(self):
        """Open the initial file."""
        self.query_collection = load_queries_from_yaml(self.current_file)

    def _new_file(self, button):
        """Create a new query collection."""
        del button
        if self._unsaved_changes() and not self.ignore_changes.value:
            print("Please save or check 'Ignore changes' before creating a new file.")
            return
        self._reset_change_state()
        self.query_collection = QueryCollection(
            file_name=_DEF_FILENAME,
            metadata=QueryMetadata(
                version=1, description="new", data_families=[], data_environments=[]
            ),
            defaults=QueryDefaults(),
            sources={"new_query": Query("New query")},
        )
        self.filename_widget.value = _DEF_FILENAME
        self._update_query_collection(self.query_collection)

    def _reset_change_state(self):
        """Reset the change state of the sub-editors."""
        self.default_param_editor.reset_changed_data()
        self.metadata_editor.reset_changed_data()
        self.query_editor.reset_changed_data()

    def _unsaved_changes(self):
        """Check if there are unsaved changes."""
        return (
            self.default_param_editor.changed_data
            or self.metadata_editor.changed_data
            or self.query_editor.changed_data
        )


# Read and write the yaml file
# Define a function to load a YAML file into a QueryCollection
def load_queries_from_yaml(yaml_file: Union[str, Path]):
    """
    Load a YAML file into a QueryCollection.

    Parameters
    ----------
    yaml_file : Union[str, Path]
        The path to the YAML file to load.

    Returns
    -------
    QueryCollection
        A QueryCollection object containing the loaded queries.

    """
    with open(yaml_file, "r", encoding="utf-8") as f_handle:
        yaml_data = yaml.safe_load(f_handle)

    metadata = QueryMetadata(**yaml_data.get("metadata", {}))
    defaults = _create_query_defaults(yaml_data.get("defaults", {}))
    queries_dict = yaml_data.get("sources", {})
    queries = {name: _create_query(query) for name, query in queries_dict.items()}
    return QueryCollection(  # type: ignore[call-arg]
        file_name=str(yaml_file), metadata=metadata, defaults=defaults, sources=queries
    )


def save_queries_to_yaml(
    query_collection: QueryCollection, yaml_file: Union[str, Path]
):
    """
    Save a QueryCollection to a YAML file.

    Parameters
    ----------
    query_collection : QueryCollection
        The QueryCollection object to save to YAML.
    yaml_file : Union[str, Path]
        The path to the YAML file to save.

    """
    query_dict = asdict(query_collection)
    # ordered_dict =
    if "file_name" in query_dict:
        del query_dict["file_name"]
    _rename_data_type(query_dict)
    yaml_data = yaml.safe_dump(_remove_none_values(query_dict), sort_keys=False)
    Path(yaml_file).write_text(yaml_data, encoding="utf-8")


def _create_query_defaults(defaults):
    """Create a QueryDefaults object."""
    def_metadata = defaults["metadata"] if "metadata" in defaults else {}
    def_params = (
        {
            name: _create_parameter(param)
            for name, param in defaults["parameters"].items()
        }
        if "parameters" in defaults and defaults["parameters"]
        else {}
    )
    return QueryDefaults(metadata=def_metadata, parameters=def_params)


def _create_query(query_data):
    """Create a Query object."""
    parameters = query_data.get("parameters", {})
    if parameters:
        parameters = {
            name: _create_parameter(param) for name, param in parameters.items()
        }
    return Query(
        description=query_data.get("description", ""),
        metadata=query_data.get("metadata", {}),
        args=QueryArgs(query=query_data.get("args", {}).get("query", "")),
        parameters=parameters,
    )


def _create_parameter(param_data):
    """Create a Parameter object."""
    return QueryParameter(
        description=param_data.get("description", ""),
        datatype=param_data.get("type", "str"),
        default=param_data.get("default"),
    )


def _remove_none_values(source_obj):
    """Recursively remove any item with a None value from a nested dictionary."""
    if isinstance(source_obj, dict):
        return {
            key: _remove_none_values(val)
            for key, val in source_obj.items()
            if val is not None or (isinstance(val, (list, dict)) and len(val) > 0)
        }
    if isinstance(source_obj, (list, tuple)):
        return type(source_obj)(
            _remove_none_values(val) for val in source_obj if val is not None
        )
    return source_obj


def _rename_data_type(source_obj):
    """Recursively rename the 'datatype' key to 'type' in a nested dictionary."""
    if isinstance(source_obj, dict):
        if "datatype" in source_obj:
            val = source_obj["datatype"]
            del source_obj["datatype"]
            source_obj["type"] = val
        for value in source_obj.values():
            _rename_data_type(value)
    if isinstance(source_obj, (list, tuple)):
        for value in source_obj:
            _rename_data_type(value)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Setup script for msticpy."""

import re

import setuptools

with open("msticpy/_version.py", "r", encoding="utf-8") as fd:
    v_match = re.search(r'^VERSION\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    __version__ = v_match[1] if v_match else "no version"

with open("requirements.txt", "r", encoding="utf-8") as fh:
    INSTALL_REQUIRES = fh.readlines()

with open("requirements-dev.txt", "r", encoding="utf-8") as fh:
    INSTALL_DEV_REQUIRES = fh.readlines()


def _combine_extras(extras: list) -> list:
    return list(
        {pkg for name, pkgs in EXTRAS.items() for pkg in pkgs if name in extras}
    )


# Extras definitions
EXTRAS = {
    "dev": INSTALL_DEV_REQUIRES,
    "vt3": ["vt-py>=0.18.0", "vt-graph-api>=2.0"],
    "splunk": ["splunk-sdk>=1.6.0,!=2.0.0"],
    "sumologic": ["sumologic-sdk>=0.1.11", "openpyxl>=3.0"],
    "kql": ["KqlmagicCustom[jupyter-extended]>=0.1.114.post22"],
    "_azure_core": [
        "azure-mgmt-compute>=4.6.2",
        "azure-mgmt-core>=1.2.1",
        "azure-mgmt-monitor>=2.0.0",
        "azure-mgmt-network>=2.7.0",
        "azure-mgmt-resource>=16.1.0",
        "azure-storage-blob>=12.5.0",
        "azure-mgmt-resourcegraph>=8.0.0",
    ],
    "azure_query": [],
    "keyvault": [],
    "ml": [
        "scikit-learn>=1.0.0",
        "scipy>=1.1.0",
        "statsmodels>=0.11.1",
        "matplotlib>=3.0.0",
    ],
    "sql2kql": ["mo-sql-parsing>=8, <9.0.0"],
    "riskiq": ["passivetotal>=2.5.3"],
    "panel": ["panel>=0.14.4"],
}
extras_all = [
    extra for name, extras in EXTRAS.items() for extra in extras if name != "dev"
]
EXTRAS["all"] = extras_all

# Create combination extras
EXTRAS["all"] = sorted(
    _combine_extras(list({name for name in EXTRAS if name != "dev"}))
)

EXTRAS["azure"] = sorted(_combine_extras(["_azure_core", "keyvault", "azure_query"]))
EXTRAS["test"] = sorted(_combine_extras(["all", "dev"]))
EXTRAS["azsentinel"] = sorted(_combine_extras(["azure", "kql", "keyvault"]))
EXTRAS["azuresentinel"] = sorted(_combine_extras(["azure", "kql", "keyvault"]))
EXTRAS["sentinel"] = sorted(_combine_extras(["azure", "kql", "keyvault"]))


if __name__ == "__main__":
    setuptools.setup(
        install_requires=INSTALL_REQUIRES, extras_require=EXTRAS, version=__version__
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data Providers Component Edit."""
import re
from typing import Optional

import ipywidgets as widgets

from .._version import VERSION
from .ce_common import TEXT_LAYOUT
from .ce_provider_base import HELP_URIS, CEProviders
from .mp_config_control import MpConfigControls

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-ancestors, duplicate-code
class CEDataProviders(CEProviders):
    """Data providers edit component."""

    _DESCRIPTION = "Settings for Data Providers"
    _COMP_PATH = "DataProviders"
    # _HELP_TEXT inherited from base
    _HELP_URI = {
        "Data Providers": (
            "https://msticpy.readthedocs.io/en/latest/" + "DataAcquisition.html"
        ),
        "Spunk": "https://msticpy.readthedocs.io/en/latest/data_acquisition/SplunkProvider.html",
        "Sumologic": (
            "https://github.com/microsoft/msticpy/blob/main/docs/notebooks/"
            "Sumologic-DataConnector.ipynb"
        ),
        "Security Datasets (aka Mordor)": (
            "https://msticpy.readthedocs.io/en/latest/data_acquisition/MordorData.html"
        ),
        **HELP_URIS,
    }

    _COMPONENT_HELP = """
    <p><b>LocalData provider <i>data_paths</i></b>
    Enter one or more data paths, separated by new lines
    </p>

    """

    def __init__(self, mp_controls: MpConfigControls):
        """
        Initialize an instance of the component.

        Parameters
        ----------
        mp_controls : MpConfigControls
            The config/controls/settings database

        """
        self.text_prov_instance = widgets.Text(
            description="Provider instance",
            placeholder="(optional) instance name for provider",
            **TEXT_LAYOUT,
        )
        super().__init__(mp_controls)
        self._last_instance_path: Optional[str] = None

    @property
    def _current_path(self):
        if self._form_current_instance_name:
            return f"{self._COMP_PATH}.{self._prov_ctrl_name}-{self._form_current_instance_name}"
        return f"{self._COMP_PATH}.{self._prov_ctrl_name}"

    @property
    def _prov_name(self) -> str:
        if self.text_prov_instance.value:
            return f"{super()._prov_name}-{self.text_prov_instance.value}"
        return super()._prov_name

    @property
    def _prov_ctrl_name(self):
        """Return the provider generic name (minus instance suffix)."""
        if "-" in super()._prov_name:
            return super()._prov_name.split("-", maxsplit=1)[0]
        return super()._prov_name

    @property
    def _select_prov_instance_name(self):
        """Return the provider instance name (minus instance suffix)."""
        if "-" in super()._prov_name:
            return super()._prov_name.split("-", maxsplit=1)[1]
        return ""

    @property
    def _form_current_instance_name(self):
        """Return the current instance name."""
        return self.text_prov_instance.value.strip()

    def _populate_edit_ctrls(
        self,
        control_name: Optional[str] = None,
        new_provider: bool = False,
    ):
        """Retrieve and populate form controls for the provider to display."""
        super()._populate_edit_ctrls(
            control_name=control_name, new_provider=new_provider
        )
        # add the instance text box
        self.edit_ctrls.children = [
            self.text_prov_instance,
            *(self.edit_ctrls.children),
        ]
        self.edit_frame.children = [self.edit_ctrls]

    def _select_provider(self, change):
        self.text_prov_instance.value = self._select_prov_instance_name
        super()._select_provider(change)
        self._last_instance_path = self._current_path

    def _save_provider(self, btn):
        if self._form_current_instance_name:
            if not re.match(r"^[\w._:]+$", self._form_current_instance_name):
                self.set_status(
                    "Error: instance name can only contain alphanumeric and '._:'"
                )
                return
            # The instance name may have changed, which alters the path
            if self._last_instance_path != self._current_path:
                self.mp_controls.rename_path(
                    self._last_instance_path, self._current_path
                )
        super()._save_provider(btn)
        # refresh the item list and re-select the current item
        edited_provider = self._prov_name
        self.select_item.options = self._get_select_opts()
        self.select_item.label = edited_provider



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Observation summary collector."""
from collections import OrderedDict
from datetime import datetime
from typing import Any, Dict, Iterator, List, Mapping, Optional, Set, Tuple

import attr
import pandas as pd
from attr import Factory
from IPython.display import Markdown, display

from .._version import VERSION
from ..common.timespan import TimeSpan

__version__ = VERSION
__author__ = "Ian Hellen"


# Query definition
@attr.s(auto_attribs=True)
class Observation:
    """
    Observation definition.

    Notes
    -----
    caption : str
        The title and index of the observation. Must
        be unique in the observation set.
    description : Optional[str]
        Text description of the observation.
        (default is None)
    data : Any
        The data to be stored for the observation (e.g.
        a pandas DataFrame). The object should implement
        a useable __repr__ to display correctly.
    data_type : Optional[str]
        The data type of the `data` property
    link : Optional[str]
        Link (usually a document-local link) to the
        originating section of the notebook.
        (default is None)
    score : int
        The risk score associated with the observation.
        (default is 0)
    tags : List[str]
        Optional list of tags.
    additional_properties Dict[str, Any]
        Additional properties not covered
        by core properties.

    """

    caption: str
    data: Any
    description: Optional[str] = None
    data_type: Optional[str] = None
    link: Optional[str] = None
    score: int = 0
    tags: List[str] = Factory(list)
    additional_properties: Dict[str, Any] = Factory(dict)
    timestamp: Optional[datetime] = None
    time_span: Optional[TimeSpan] = None
    time_column: Optional[str] = None
    filter: Optional[str] = None
    schema: Optional[str] = None

    @classmethod
    def required_fields(cls) -> List[str]:
        """
        Return required fields for Observation instance.

        Returns
        -------
        List[str]
            List of field names.

        """
        return ["caption", "data"]

    @classmethod
    def all_fields(cls) -> Set[str]:
        """
        Return all fields of Observation class.

        Returns
        -------
        Set[str]
            Set of all field names.

        """
        return {field.name for field in attr.fields(cls)}  # type: ignore[misc]

    def display(self):
        """Display the observation."""
        display(Markdown(f"### {self.caption}"))
        if self.description:
            display(Markdown(self.description))
        if self.score:
            display(Markdown(f"Score: {self.score}"))
        if self.link:
            display(Markdown(f"[Go to details](#{self.link})"))
        if self.tags:
            display(Markdown(f'tags: {", ".join(self.tags)}'))
        display(self.filtered_data)
        if self.additional_properties:
            display(Markdown("### Additional Properties"))
            # pylint: disable=no-member
            for key, val in self.additional_properties.items():
                display(Markdown(f"**{key}**: {val}"))
            # pylint: enable=no-member

    @property
    def filtered_data(self) -> Any:
        """Apply filtering to data if it is a DataFrame."""
        if not isinstance(self.data, pd.DataFrame):
            return self.data
        filtered_data = self.data
        if self.filter:
            filtered_data = filtered_data.query(self.filter)
        if self.time_span and self.time_column:
            filtered_data = filtered_data[
                (filtered_data[self.time_column] >= self.time_span.start)
                & (filtered_data[self.time_column] <= self.time_span.end)
            ]
        return filtered_data


class Observations:
    """Class to collect and display investigation observations."""

    def __init__(self, observationlist: "Observations" = None):
        """
        Create an observation list.

        Parameters
        ----------
        observationlist : Observations, optional
            Initialize from an existing Observations list
            (the default is None)

        """
        self.observation_list: Dict[str, Observation] = OrderedDict()
        if observationlist is not None:
            self.observation_list.update(observationlist.observations)

    def __getitem__(self, key: str) -> Observation:
        """Return the observation with a caption."""
        return self.observation_list[key]

    def __iter__(self) -> Iterator[Tuple[str, Observation]]:
        """Return iterator over observations."""
        yield from self.observation_list.items()

    @property
    def observations(self) -> Mapping[str, Observation]:
        """
        Return the current list of Observations.

        Returns
        -------
        Mapping[str, Observation]
            The current ordered dictionary of Observations

        """
        return self.observation_list

    def display_observations(self):
        """Display the current observations using IPython.display."""
        for observation in self.observation_list.values():
            display(observation)

    def add_observation(self, observation: Observation = None, **kwargs):
        """
        Add an observation.

        Add an observation as an Observation instance or as a
        set of keyword parameters (see Observation class for
        acceptable values). Any keyword parameters that are not
        properties of Observation will be stored in the
        Observation.additional_properties dictionary

        Parameters
        ----------
        observation : Observation
            An observation instance.

        Other Parameters
        ----------------
        kwargs : str, Any
            List of key value pairs of the property names
            and values of the Observation to be stored.

        """
        if observation is not None:
            self.observation_list[observation.caption] = observation
        else:
            req_fields = set(Observation.required_fields())
            missing_fields = req_fields.difference(kwargs.keys())
            if missing_fields:
                raise ValueError(
                    "The following fields are required",
                    f"in an Observation: {missing_fields}",
                )

            core_fields = {
                key: value
                for key, value in kwargs.items()
                if key in Observation.all_fields()
            }
            new_observation = Observation(**core_fields)
            addl_fields = {
                key: value
                for key, value in kwargs.items()
                if key not in Observation.all_fields()
            }
            # pylint: disable=no-member
            new_observation.additional_properties.update(addl_fields)
            self.observation_list[new_observation.caption] = new_observation



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSAL delegated app authentication class."""
import json
import logging
from sys import platform

import msal
from msal_extensions import (
    FilePersistence,
    FilePersistenceWithDataProtection,
    KeychainPersistence,
    LibsecretPersistence,
    PersistedTokenCache,
)
from msal_extensions.persistence import PersistenceNotFound

from .._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"


class MSALDelegatedAuth:
    """Class for handling MSAL based delegated auth."""

    def __init__(
        self,
        client_id: str,
        authority: str,
        username: str,
        scopes: list,
        debug: bool = False,
        **kwargs,
    ):
        """
        Instantiate MSAL authentication class.

        Parameters
        ----------
        client_id : str
            ID of the client to connect to.
        authority : str
            The authentication authority to authenticate against.
        username : str
            The username to authenticate with.
        scopes : list
            A list of scopes to authenicate with
        debug : bool, optional
            Set True to get additional logging ouput, by default False

        """
        self.token_cache = None
        self.location = kwargs.get("location", "token_cache.bin")
        self.auth_type = kwargs.get("auth_type", "interactive")
        self.username = username
        self.scopes = scopes
        self.result = None

        persistence = self._create_cache()
        if persistence:
            self.token_cache = PersistedTokenCache(persistence)

        self.app = msal.PublicClientApplication(
            client_id=client_id, authority=authority, token_cache=self.token_cache
        )

        if "connect" in kwargs:
            self.get_token()

        if not debug:
            logging.getLogger("msal").setLevel(logging.ERROR)

    def get_token(self):
        """Get an authneticaiton token."""
        chosen_account = self.app.get_accounts(username=self.username)
        if chosen_account:
            self.result = self.app.acquire_token_silent_with_error(
                scopes=self.scopes, account=chosen_account[0]
            )
            if not self.result:
                self.result = self._app_auth(self.auth_type)
        else:
            self.result = self._app_auth(self.auth_type)
        self.refresh_token()

    def refresh_token(self):
        """Refresh the authentication token."""
        self.result = None
        chosen_account = self.app.get_accounts(username=self.username)
        if chosen_account:
            self.result = self.app.acquire_token_silent_with_error(
                scopes=self.scopes, account=chosen_account[0], force_refresh=True
            )
        if not self.result:
            self.get_token()

    def _app_auth(self, auth_type):  # pylint: disable=inconsistent-return-statements
        """Conduct authentication against the application."""
        if auth_type == "interactive":
            return self.app.acquire_token_interactive(scopes=self.scopes)
        if auth_type == "device":
            flow = self.app.initiate_device_flow(scopes=self.scopes)
            if "user_code" not in flow:
                raise ValueError(
                    f"Fail to create device flow. Err: {json.dumps(flow, indent=4)}"
                )
            print(flow["message"])
            return self.app.acquire_token_by_device_flow(flow)

    def _create_cache(self):
        """Build a suitable token persistence instance based your current OS."""
        if platform.startswith("win"):
            return FilePersistenceWithDataProtection(self.location)
        if platform.startswith("darwin"):
            return KeychainPersistence(self.location, "msal_token", "msal_token_value")
        if platform.startswith("linux"):
            try:
                return LibsecretPersistence(
                    self.location,
                    schema_name="msal_token",
                    attributes={
                        "msal_token1": "msal_token_values",
                        "msal_token2": "msal_token_values",
                    },
                )
            except (PersistenceNotFound, ImportError, ValueError):
                print("Unable to create encrypted token cache - using in memory cache.")
                return None

        return FilePersistence(self.location)

    @property
    def token(self):
        """Return the token."""
        return self.result["access_token"]
